#ifndef FSGUI_DVIEWCONTROLDIALOG_IS_INCLUDED
#define FSGUI_DVIEWCONTROLDIALOG_IS_INCLUDED
/* { */

#include "fsgui3dviewcontrol.h"
#include <fsgui.h>


class FsGui3DViewControlDialogBase : public FsGuiDialog
{
protected:
	FsGuiCanvas *canvas;
	class FsGui3DViewControl *viewControl;
	friend class FsGui3DViewControl;

	class PosAttZoomDist
	{
	public:
		YsVec3 pos;
		YsAtt3 att;
		double zoom,dist;
	};
	PosAttZoomDist lastView;



protected:
	virtual ~FsGui3DViewControlDialogBase();

public:
	/*! Deletes a dialog object. */
	static void Delete(FsGui3DViewControlDialogBase *toDel);

	/*! This function returns a reference to the internally-stored viewing-control information. */
	FsGui3DViewControl &GetViewControl(void);

	/*! This function returns a const reference to the internally-stored viewing-control information. */
	const FsGui3DViewControl &GetViewControl(void) const;

public:
	/*! This function saves the current view point as last view so that the user can recall the view point later. */
	void SaveLastView(void);

protected:
	PosAttZoomDist RecallLastView(void);

public:
	virtual YSBOOL SingleButtonRotation(void) const=0;
	virtual YSBOOL SingleButtonScroll(void) const=0;
	virtual YSBOOL SingleButtonZoom(void) const=0;
	virtual void SetOrientationLock(YSBOOL lock)=0;
	virtual void SetTranslationLock(YSBOOL lock)=0;
	virtual void UpdateViewTarget(const YsVec3 &pos)=0;
};



class FsGui3DViewControlDialog : public FsGui3DViewControlDialogBase
{
private:
	FsGuiButton *openBtn,*closeBtn;

	FsGuiButton *rollLeftBtn,*noseDownBtn,*rollRightBtn;
	FsGuiButton *yawLeftBtn,*zeroAttBtn,*yawRightBtn;
	FsGuiButton *left90Btn,*noseUpBtn,*right90Btn;

	FsGuiButton *scrollUpBtn;
	FsGuiButton *scrollLeftBtn,*centerBtn,*scrollRightBtn;
	FsGuiButton *scrollDownBtn;

	FsGuiButton *lockOrientationBtn;
	FsGuiButton *lockTranslationBtn;

	FsGuiButton *frontViewBtn,*rearViewBtn;
	FsGuiButton *topViewBtn,*bottomViewBtn;
	FsGuiButton *leftViewBtn,*rightViewBtn;

	FsGuiButton *zoomBtn,*moozBtn;
	FsGuiButton *orthoPersBtn;

	FsGuiTextBox *lookAtCoordTxt;
	FsGuiButton *lookAtBtn,*lastViewBtn;
	FsGuiButton *lastLookAtBtn;
	FsGuiButton *lookAtCutBufBtn;

	FsGuiButton *copyBtn,*pasteBtn;

	FsGuiButton *noSingleBtnOpBtn,*rotateBtn,*translateBtn;

	FsGuiButton *saveViewBtn,*loadViewBtn;



private:
	FsGui3DViewControlDialog();
	virtual ~FsGui3DViewControlDialog();

public:
	/*! Creates a dialog object. */
	static FsGui3DViewControlDialog *Create(void);

	/*! This function makes the dialog from a view controller. */
	void Make(class FsGuiCanvas *canvas,class FsGui3DViewControl *viewControl,YSBOOL minimize=YSFALSE);

	/*! Call back function for button click. */
	virtual void OnButtonClick(FsGuiButton *btn);

	/*! This function sets lock-orientation status. */
	virtual void SetOrientationLock(YSBOOL lock);

	/*! This function sets lock-translation status. */
	virtual void SetTranslationLock(YSBOOL lock);

	/*! This function returns the status of "Rotate" button for single-button operation. */
	virtual YSBOOL SingleButtonRotation(void) const;

	/*! This function returns the status of "Scroll" button for single-button operation. */
	virtual YSBOOL SingleButtonScroll(void) const;

	/*! This function returns the status of "SingleButtonZoom" button.  Not included in the standard-view control dialog. */
	virtual YSBOOL SingleButtonZoom(void) const;

private:
	void Translate(double vx,double vy);
	virtual void UpdateViewTarget(const YsVec3 &pos);
};

/* } */
#endif
