#include "fsgui3dapp.h"

#include <ysgl.h>

static YSBOOL appMustTerminate=YSFALSE;
static int appSleepInterval=10;
static FsGui3DMainCanvas *appMainCanvas;

////////////////////////////////////////////////////////////

FsGui3DMainMenu::FsGui3DMainMenu(class FsGui3DMainCanvas *canvas)
{
	this->canvas=canvas;
}

void FsGui3DMainMenu::OnSelectMenuItem(FsGuiPopUpMenuItem *item)
{
	canvas->OnSelectMenuItem(item);
}

////////////////////////////////////////////////////////////

FsGui3DMainCanvas::FsGui3DMainCanvas() : threeDInterface(this),viewControlDlg(NULL)
{
	MakeMainMenu();
	AddViewControlDialog();
	drawEnv.AttachCanvas(this);
	menuIdSource=0x1000000;
}

FsGui3DMainCanvas::~FsGui3DMainCanvas()
{
	// The following two lines ensure that all self-destructive dialogs are cleaned. 2015/03/18
	RemoveDialogAll();
	PerformScheduledDeletion();

	DeleteMainMenu();
	if(NULL!=viewControlDlg)
	{
		FsGui3DViewControlDialog::Delete(viewControlDlg);
	}
}

FsGuiPopUpMenuItem *FsGui3DMainCanvas::AddMenuItem(FsGuiPopUpMenu *parent,int fsKey,const wchar_t label[],void (THISCLASS::*funcPtr)(FsGuiPopUpMenuItem *))
{
	auto menuPtr=parent->AddTextItem(menuIdSource,fsKey,label);
	menuConn.Register(menuIdSource,this,funcPtr);
	menuIdSource++;
	return menuPtr;
}

FsGuiPopUpMenuItem *FsGui3DMainCanvas::AddMenuItem(FsGuiPopUpMenu *parent,int fsKey,const char label[],void (THISCLASS::*funcPtr)(FsGuiPopUpMenuItem *))
{
	auto menuPtr=parent->AddTextItem(menuIdSource,fsKey,label);
	menuConn.Register(menuIdSource,this,funcPtr);
	menuIdSource++;
	return menuPtr;
}

void FsGui3DMainCanvas::RegisterLButtonDownCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int))
{
	LButtonDownCallBack.Register(this,funcPtr);
}
void FsGui3DMainCanvas::RegisterLButtonUpCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int))
{
	LButtonUpCallBack.Register(this,funcPtr);
}
void FsGui3DMainCanvas::RegisterMButtonDownCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int))
{
	MButtonDownCallBack.Register(this,funcPtr);
}
void FsGui3DMainCanvas::RegisterMButtonUpCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int))
{
	MButtonUpCallBack.Register(this,funcPtr);
}
void FsGui3DMainCanvas::RegisterRButtonDownCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int))
{
	RButtonDownCallBack.Register(this,funcPtr);
}
void FsGui3DMainCanvas::RegisterRButtonUpCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int))
{
	RButtonUpCallBack.Register(this,funcPtr);
}

void FsGui3DMainCanvas::RegisterMouseMoveCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int,int,int))
{
	MouseMoveCallBack.Register(this,funcPtr);
}

void FsGui3DMainCanvas::RegisterCloseModalCallBack(void (THISCLASS::*funcPtr)(FsGuiDialog *,int))
{
	closeModalCallBack.Register(this,funcPtr);
}

void FsGui3DMainCanvas::AddViewControlDialog(void)
{
	viewControlDlg=FsGui3DViewControlDialog::Create();
	viewControlDlg->Make(this,&drawEnv);
	this->AddDialog(viewControlDlg);
}

void FsGui3DMainCanvas::MakeMainMenu(void)
{
	menuConn.CleanUp();

	mainMenu=new FsGui3DMainMenu(this);
	mainMenu->Initialize();
	mainMenu->SetIsPullDownMenu(YSTRUE);

	FsGuiPopUpMenuItem *fileMenu=mainMenu->AddTextItem(0,FSKEY_F,L"File");
	FsGuiPopUpMenu *fileSubMenu=fileMenu->GetSubMenu();

	AddMenuItem(fileSubMenu,FSKEY_X,L"Exit",&THISCLASS::File_Exit);

	SetMainMenu(mainMenu);
}

void FsGui3DMainCanvas::DeleteMainMenu(void)
{
	delete mainMenu;
}

void FsGui3DMainCanvas::OnInterval(void)
{
	FsGuiCanvas::Interval();

	{
		int key;
		while(FSKEY_NULL!=(key=FsInkey()))
		{
			if(YSOK!=drawEnv.KeyIn(key,(YSBOOL)FsGetKeyState(FSKEY_SHIFT),(YSBOOL)FsGetKeyState(FSKEY_CTRL),(YSBOOL)FsGetKeyState(FSKEY_ALT)))
			{
				if(YSOK!=threeDInterface.KeyIn(key,(YSBOOL)FsGetKeyState(FSKEY_SHIFT),(YSBOOL)FsGetKeyState(FSKEY_CTRL),(YSBOOL)FsGetKeyState(FSKEY_ALT)))
				{
					this->KeyIn(key,(YSBOOL)FsGetKeyState(FSKEY_SHIFT),(YSBOOL)FsGetKeyState(FSKEY_CTRL),(YSBOOL)FsGetKeyState(FSKEY_ALT));
				}
			}
		}
	}

	{
		int charCode;
		while(0!=(charCode=FsInkeyChar()))
		{
			this->CharIn(charCode);
		}
	}

	{
		int lb,mb,rb,mx,my;
		while(FSMOUSEEVENT_NONE!=FsGetMouseEvent(lb,mb,rb,mx,my))
		{
			if(YSOK!=drawEnv.SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my))
			{
				if(YSOK!=this->SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my))
				{
					if(YSOK!=threeDInterface.SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my))
					{
					}
				}
			}
		}
	}

	int winWid,winHei;
	FsGetWindowSize(winWid,winHei);
	this->SetWindowSize(winWid,winHei,/*autoArrangeDialog=*/YSTRUE);

	drawEnv.SetWindowSize(winWid,winHei);
	drawEnv.SetViewportByTwoCorner(0,0,winWid,winHei);


	if(YSTRUE==NeedRedraw() ||
	   YSTRUE==drawEnv.NeedRedraw() ||
	   YSTRUE==threeDInterface.NeedRedraw() ||
	   0!=FsCheckWindowExposure())
	{
		Draw();
	}
}

void FsGui3DMainCanvas::OnSelectMenuItem(FsGuiPopUpMenuItem *menuItem)
{
	if(NULL!=menuItem)
	{
		menuConn.Invoke(menuItem);
	}
}

YSRESULT FsGui3DMainCanvas::OnKeyDown(int fskey,YSBOOL shift,YSBOOL ctrl,YSBOOL alt)
{
	if(YSOK==menuConn.InvokeFromShortCut(fskey,shift,ctrl,alt))
	{
		return YSOK;
	}
	return YSERR;
}

YSRESULT FsGui3DMainCanvas::OnLButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my)
{
	return LButtonDownCallBack.Invoke(lb,mb,rb,mx,my);
}
YSRESULT FsGui3DMainCanvas::OnMButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my)
{
	return MButtonDownCallBack.Invoke(lb,mb,rb,mx,my);
}
YSRESULT FsGui3DMainCanvas::OnRButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my)
{
	return RButtonDownCallBack.Invoke(lb,mb,rb,mx,my);
}
YSRESULT FsGui3DMainCanvas::OnLButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my)
{
	return LButtonUpCallBack.Invoke(lb,mb,rb,mx,my);
}
YSRESULT FsGui3DMainCanvas::OnMButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my)
{
	return MButtonUpCallBack.Invoke(lb,mb,rb,mx,my);
}
YSRESULT FsGui3DMainCanvas::OnRButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my)
{
	return RButtonUpCallBack.Invoke(lb,mb,rb,mx,my);
}

YSRESULT FsGui3DMainCanvas::OnMouseMove(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my,int prevMx,int prevMy)
{
	return MouseMoveCallBack.Invoke(lb,mb,rb,mx,my,prevMx,prevMy);
}

/* virtual */ void FsGui3DMainCanvas::OnModalDialogClosed(int dialogIdent,class FsGuiDialog *closedModalDialog,int modalDialogEndCode)
{
	auto callbackCopy=closeModalCallBack;  // Caution!  closeModalCallBack may be updated in Invoke.
	closeModalCallBack.Register(NULL,NULL);  // Therefore, clearing call back must be done before Invoke.
	callbackCopy.Invoke(closedModalDialog,modalDialogEndCode);  // And, it should be Invoked from the copy.
}

void FsGui3DMainCanvas::Draw(void)
{
	// Do this at the beginning of Draw funtion.  This will allow one of the elements set SetNeedRedraw(YSTRUE) 
	// within drawing function so that Draw function will be called again in the next iteragion. >>
	SetNeedRedraw(YSFALSE);
	drawEnv.SetNeedRedraw(YSFALSE);
	threeDInterface.SetNeedRedraw(YSFALSE);
	// <<

	int viewport[4];
	drawEnv.GetOpenGlCompatibleViewport(viewport);
	drawEnv.SetVerticalOrientation(YSTRUE);
	drawEnv.UpdateNearFar();


	glUseProgram(0);


	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);


	const YsMatrix4x4 &projMat=drawEnv.GetProjectionMatrix();
	const YsMatrix4x4 &viewMat=drawEnv.GetViewMatrix();
	const YsMatrix4x4 projViewMat=(projMat*viewMat);

	threeDInterface.SetViewport(viewport);
	threeDInterface.SetProjViewModelMatrix(projViewMat);


	glViewport(viewport[0],viewport[1],viewport[2],viewport[3]);

	GLfloat glProjMat[16];
	projMat.GetOpenGlCompatibleMatrix(glProjMat);

	GLfloat glModelviewMat[16];
	viewMat.GetOpenGlCompatibleMatrix(glModelviewMat);


	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMultMatrixf(glProjMat);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMultMatrixf(glModelviewMat);


	YsGLSLSetShared3DRendererProjection(glProjMat);
	YsGLSLSetShared3DRendererModelView(glModelviewMat);



	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	threeDInterface.Draw();


	glUseProgram(0);
	glColor3ub(0,0,0);
	glBegin(GL_LINES);
	glVertex3d(0.0,0.0,0.0);
	glVertex3d(1.0,0.0,0.0);
	glVertex3d(0.0,0.0,0.0);
	glVertex3d(0.0,1.0,0.0);
	glVertex3d(0.0,0.0,0.0);
	glVertex3d(0.0,0.0,1.0);
	glEnd();


	int nTri=0;
	YsVec3 tri[9];
	if(YSOK==FsGui3DGetArrowHeadTriangle(
	    tri+nTri*3,YsOrigin(),YsXVec(),viewport,projMat,viewMat,0.1,10,8))
	{
		++nTri;
	}
	if(YSOK==FsGui3DGetArrowHeadTriangle(
	    tri+nTri*3,YsOrigin(),YsYVec(),viewport,projMat,viewMat,0.1,10,8))
	{
		++nTri;
	}
	if(YSOK==FsGui3DGetArrowHeadTriangle(
	    tri+nTri*3,YsOrigin(),YsZVec(),viewport,projMat,viewMat,0.1,10,8))
	{
		++nTri;
	}

	if(0<nTri)
	{
		glBegin(GL_TRIANGLES);
		for(int i=0; i<nTri*3; ++i)
		{
			glVertex3dv(tri[i]);
		}
		glEnd();
	}


	// 2D Drawing
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0,(double)viewport[2],(double)viewport[3],0.0,-1.0,1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	YsGLSLUsePlain2DRenderer(YsGLSLSharedPlain2DRenderer());
	YsGLSLUseWindowCoordinateInPlain2DDrawing(YsGLSLSharedPlain2DRenderer(),YSTRUE);
	YsGLSLEndUsePlain2DRenderer(YsGLSLSharedPlain2DRenderer());

	glDisable(GL_DEPTH_TEST);

	glUseProgram(0);
	FsGuiCanvas::Show();

	FsSwapBuffers();
}



////////////////////////////////////////////////////////////

void FsGui3DMainCanvas::File_Exit(FsGuiPopUpMenuItem *)
{
	auto msgDlg=FsGuiDialog::CreateSelfDestructiveDialog <FsGuiMessageBoxDialog>();
	msgDlg->Make(L"Confirm Exit?",L"Confirm Exit?",L"Yes",L"No");
	AttachModalDialog(msgDlg);
	RegisterCloseModalCallBack(&THISCLASS::File_Exit_ConfirmExitCallBack);
}

void FsGui3DMainCanvas::File_Exit_ConfirmExitCallBack(FsGuiDialog *,int returnValue)
{
	if(YSTRUE==(YSBOOL)returnValue)
	{
		File_Exit_ReallyExit();
	}
}

void FsGui3DMainCanvas::File_Exit_ReallyExit(void)
{
	appMustTerminate=YSTRUE;
}

////////////////////////////////////////////////////////////


YSBOOL FsGui3dAppMustTerminate(void)
{
	return appMustTerminate;
}

int FsGui3dAppSleepIntervalInMS(void)
{
	return appSleepInterval;
}

static void FsGui3DAppLocalIntervalHook(void *)
{
	FsGui3DAppIntervalCallBack();
}

void FsGui3DAppInitialize(void)
{
	appMainCanvas=new FsGui3DMainCanvas;
	YsGLSLCreateSharedRenderer();

	appMainCanvas->drawEnv.SetTargetBoundingBox(-5.0*YsXYZ(),5.0*YsXYZ());
	appMainCanvas->drawEnv.SetViewTarget(YsOrigin());
	appMainCanvas->drawEnv.SetViewDistance(7.5);

	FsRegisterIntervalCallBack(FsGui3DAppLocalIntervalHook,NULL);
}

void FsGui3DAppIntervalCallBack(void)
{
	appMainCanvas->OnInterval();
}

void FsGui3dAppCleanUp(void)
{
	YsGLSLDeleteSharedRenderer();
	delete appMainCanvas;
}

