#ifndef FSGUI_DAPP_IS_INCLUDED
#define FSGUI_DAPP_IS_INCLUDED
/* { */

#include <fsgui3d.h>

// Call back functions from the main framework. >>
YSBOOL FsGui3dAppMustTerminate(void);
int FsGui3dAppSleepIntervalInMS(void);

void FsGui3DAppInitialize(void);
void FsGui3DAppIntervalCallBack(void);
void FsGui3dAppCleanUp(void);
// Call back functions from the main framework. <<

class FsGui3DMainMenu : public FsGuiPopUpMenu
{
private:
	class FsGui3DMainCanvas *canvas;
public:
	FsGui3DMainMenu(class FsGui3DMainCanvas *canvas);
	virtual void OnSelectMenuItem(FsGuiPopUpMenuItem *item);
};

class FsGui3DMainCanvas : public FsGuiCanvas
{
public:
	enum
	{
		MENU_FILE_EXIT=0x7f000001
	};

	FsGui3DInterface threeDInterface;
	FsGui3DViewControl drawEnv;
	FsGui3DViewControlDialog *viewControlDlg;
	FsGui3DMainMenu *mainMenu;



private:
	// Self-memo: 
	//   It is obvious after realizing that the regular member function is unique to this class, not the parent class.
	//   But the question is can I make this menu-function call-back as a feature of FsGuiCanvas?
	//   The answer is no.  THISCLASS must be FsGui3DMainCanvas, which is the actual implementation, not FsGuiCanvas.
	//   Therefore, it looks to be better to make it a part of the application template.
	typedef FsGui3DMainCanvas THISCLASS;
	int menuIdSource;
	FsGuiMenuCallBackConnector <THISCLASS> menuConn;
	FsGuiPopUpMenuItem *AddMenuItem(FsGuiPopUpMenu *parent,int fsKey,const char label[],void (THISCLASS::*funcPtr)(FsGuiPopUpMenuItem *));
	FsGuiPopUpMenuItem *AddMenuItem(FsGuiPopUpMenu *parent,int fsKey,const wchar_t label[],void (THISCLASS::*funcPtr)(FsGuiPopUpMenuItem *));

	FsGuiMouseButtonCallBackFunction <THISCLASS> LButtonDownCallBack,LButtonUpCallBack,MButtonDownCallBack,MButtonUpCallBack,RButtonDownCallBack,RButtonUpCallBack;
	void RegisterLButtonDownCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int));
	void RegisterLButtonUpCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int));
	void RegisterMButtonDownCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int));
	void RegisterMButtonUpCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int));
	void RegisterRButtonDownCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int));
	void RegisterRButtonUpCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int));

	FsGuiMouseMoveCallBackFunction <THISCLASS> MouseMoveCallBack;
	void RegisterMouseMoveCallBack(YSRESULT (THISCLASS::*funcPtr)(YSBOOL,YSBOOL,YSBOOL,int,int,int,int));

	FsGuiDialogCallBackFunction <THISCLASS,int> closeModalCallBack;
	void RegisterCloseModalCallBack(void (THISCLASS::*funcPtr)(FsGuiDialog *,int));

public:
	// [Core data structure]


	// [Modeless dialogs]


	// [Modal dialogs]
	//   (1) Add a pointer in the following chunk.
	//   (2) Add an initialization in the constructor of the aplication.
	//   (3) Add deletion in the destructor of the application.
	//   (4) Add RemoveDialog in Edit_ClearUIIMode


	FsGui3DMainCanvas();
	~FsGui3DMainCanvas();
private:
	void AddViewControlDialog(void);
	void MakeMainMenu(void);
	void DeleteMainMenu(void);

public:
	void OnInterval(void);
	void OnSelectMenuItem(FsGuiPopUpMenuItem *menuItem);
	YSRESULT OnKeyDown(int fskey,YSBOOL shift,YSBOOL ctrl,YSBOOL alt);
	YSRESULT OnLButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnMButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnRButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnLButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnMButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnRButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnMouseMove(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my,int prevMx,int prevMy);
	virtual void OnModalDialogClosed(int dialogIdent,class FsGuiDialog *closedModalDialog,int modalDialogEndCode);

	void Draw(void);

private:
	// [Menu pointers for check marks]


	// [Menu call-backs]
	void File_Exit(FsGuiPopUpMenuItem *);
	void File_Exit_ConfirmExitCallBack(FsGuiDialog *,int);
	void File_Exit_ReallyExit(void);
};



/* } */
#endif
