#include "fsguibase.h"
#include "fsguidialog.h"

FsGuiObject::SCHEME FsGuiObject::scheme=FsGuiObject::PLAIN;

int FsGuiObject::defBitmapWidthLimit=64;
int FsGuiObject::defBitmapHeightLimit=64;


FsGuiControlBase::FsGuiControlBase()
{
	parent=NULL;
	modalDlg=NULL;
}


/* virtual */ FsGuiControlBase::~FsGuiControlBase()
{
	PerformScheduledDeletion();
}


void FsGuiControlBase::SetParent(FsGuiControlBase *parent)
{
	this->parent=parent;
}


FsGuiControlBase *FsGuiControlBase::GetParent(void) const
{
	return parent;
}

void FsGuiControlBase::AttachModalDialog(class FsGuiDialog *modalDlg)
{
	this->modalDlg=modalDlg;
	if(NULL!=modalDlg)
	{
		modalDlg->SetParent(this);
		modalDlg->OnAttach();
	}
	SetNeedRedraw(YSTRUE);
}

void FsGuiControlBase::DetachModalDialog(void)
{
	if(NULL!=modalDlg)
	{
		auto toCall=modalDlg;
		AttachModalDialog(NULL);
		toCall->OnDetach();

		if(YSTRUE==toCall->IsSelfDestructiveDialog())
		{
			scheduledForDeletion.Append(toCall);
		}
	}
	else
	{
		AttachModalDialog(NULL);
	}
	SetNeedRedraw(YSTRUE);
}

class FsGuiDialog *FsGuiControlBase::GetModalDialog(void) const
{
	return modalDlg;
}

void FsGuiControlBase::Interval(void)
{
}

void FsGuiControlBase::OnModalDialogClosed(int,class FsGuiDialog *,int)
{
}

YSRESULT FsGuiControlBase::OnSaveView(const class FsGui3DViewControlDialog *dlg)
{
	return YSERR;
}

YSRESULT FsGuiControlBase::OnLoadView(class FsGui3DViewControlDialog *dlg)
{
	return YSERR;
}

////////////////////////////////////////////////////////////

YSBOOL FsGuiBitmapType::IsEmpty(void) const
{
	if(0==GetWidth() || 0==GetHeight())
	{
		return YSTRUE;
	}
	return YSFALSE;
}

YSBOOL FsGuiBitmapType::IsLoaded(void) const
{
	if(0==GetWidth() || 0==GetHeight())
	{
		return YSFALSE;
	}
	return YSTRUE;
}

int FsGuiBitmapType::GetWidth(void) const
{
	return GetTotalWidth();
}

int FsGuiBitmapType::GetHeight(void) const
{
	return GetTotalHeight();
}

void FsGuiBitmapType::MakeSmallEmptyBitmap(void)
{
	YsBitmap bmp;
	bmp.PrepareBitmap(8,8);
	SetAndBreakUp(bmp,8,8);
}

void FsGuiBitmapType::SetBitmap(const YsBitmap &bmpIn)
{
	SetAndBreakUp(bmpIn,FsGuiObject::defBitmapWidthLimit,FsGuiObject::defBitmapHeightLimit);
}

void FsGuiControlBase::PerformScheduledDeletion(void)
{
	for(auto ptr : scheduledForDeletion)
	{
		delete ptr;
	}
	if(0<scheduledForDeletion.GetN())
	{
		YsPrintf("Deleted %d dialogs scheduled for deletion.\n",(int)scheduledForDeletion.GetN());
	}
	scheduledForDeletion.CleanUp();
}
