#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>

#include "fssimplewindow.h"

extern "C" void FsOpenWindowC(int x0,int y0,int wid,int hei,int useDoubleBuffer,int useMultiSampleBuffer,const char windowName[]);
extern "C" void FsGetWindowSizeC(int *wid,int *hei);
extern "C" void FsGetWindowPositionC(int *wid,int *hei);
extern "C" void FsMaximizeWindowC(void);
extern "C" void FsPollDeviceC(void);
extern "C" void FsPushOnPaintEventC(void);
extern "C" void FsSleepC(int ms);
extern "C" int FsPassedTimeC(void);
extern "C" long long int FsSubSecondTimerC(void);
extern "C" void FsMouseC(int *lb,int *mb,int *rb,int *mx,int *my);
extern "C" int FsGetMouseEventC(int *lb,int *mb,int *rb,int *mx,int *my);
extern "C" void FsSwapBufferC(void);
extern "C" int FsInkeyC(void);
extern "C" int FsInkeyCharC(void);
extern "C" int FsKeyStateC(int fsKeyCode);
extern "C" void FsChangeToProgramDirC(void);
extern "C" int FsCheckExposureC(void);



static int doubleBuffer=1;
static int windowOpen=0;

void FsOpenWindow(const FsOpenWindowOption &opt)
{
	if(0!=windowOpen)
	{
		printf("Error! Window is already open.\n");
		exit(1);
	}

	const int useDoubleBuffer=(int)opt.useDoubleBuffer;
	const int useMultiSampleBuffer=(int)opt.useMultiSampleBuffer;

	if(NULL==opt.windowTitle)
	{
		FsOpenWindowC(opt.x0,opt.y0,opt.wid,opt.hei,useDoubleBuffer,useMultiSampleBuffer,"Main Window");
	}
	else
	{
		FsOpenWindowC(opt.x0,opt.y0,opt.wid,opt.hei,useDoubleBuffer,useMultiSampleBuffer,opt.windowTitle);
	}

	FsPassedTimeC();  // Resetting the timer.
	doubleBuffer=useDoubleBuffer;
	windowOpen=1;

	if(NULL!=fsAfterWindowCreationCallBack)
	{
		(*fsAfterWindowCreationCallBack)(fsAfterWindowCreationCallBackParam);
	}
}

void FsPushOnPaintEvent(void)
{
	FsPushOnPaintEventC();
}

void FsCloseWindow(void)
{
}

void FsGetWindowSize(int &wid,int &hei)
{
	FsGetWindowSizeC(&wid,&hei);
}

void FsGetWindowPosition(int &x0,int &y0)
{
	FsGetWindowPositionC(&x0,&y0);
}

void FsSetWindowTitle(const char windowTitle[])
{
	printf("Sorry. %s not supported on this platform yet\n",__FUNCTION__);
}

void FsMaximizeWindow(void)
{
	FsMaximizeWindowC();
}

void FsPollDevice(void)
{
	FsPollDeviceC();
}

void FsSleep(int ms)
{
	FsSleepC(ms);
}

long long int FsPassedTime(void)
{
	static long long int lastTick;
	long long int tick;

	static int first=1;
	if(1==first)
	{
		lastTick=FsSubSecondTimer();
		first=0;
	}

	tick=FsSubSecondTimer();
	long long passed=tick-lastTick;
	lastTick=tick;

	return passed;
}

long long int FsSubSecondTimer(void)
{
	return FsSubSecondTimerC();
}

void FsGetMouseState(int &lb,int &mb,int &rb,int &mx,int &my)
{
	FsMouseC(&lb,&mb,&rb,&mx,&my);
}

int FsGetMouseEvent(int &lb,int &mb,int &rb,int &mx,int &my)
{
	return FsGetMouseEventC(&lb,&mb,&rb,&mx,&my);
}

void FsSwapBuffers(void)
{
	if(0==doubleBuffer)
	{
		printf("Error! FsSwapBufferes called in a single-buffered application.\n");
		exit(1);
	}

	FsSwapBufferC();
}

int FsInkey(void)
{
	return FsInkeyC();
}

int FsInkeyChar(void)
{
	return FsInkeyCharC();
}

int FsGetKeyState(int fsKeyCode)
{
	return FsKeyStateC(fsKeyCode);
}

int FsCheckWindowExposure(void)
{
	return FsCheckExposureC();
}

void FsChangeToProgramDir(void)
{
	FsChangeToProgramDirC();
}

extern "C" void FsOnPaintCallBackCpp(void)
{
	if(NULL!=fsOnPaintCallback)
	{
		(*fsOnPaintCallback)(fsOnPaintCallbackParam);
	}
}
