#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stddef.h>
#include <string.h>

#ifdef _WIN32
	#include <windows.h>
	#include <direct.h>
#endif

#include "fssimplewindow.h"





class FsMouseEventLog
{
public:
	int eventType;
	int lb,mb,rb;
	int mx,my;
	unsigned int shift,ctrl;
};


#define NKEYBUF 256
static int keyBuffer[NKEYBUF];
static int nKeyBufUsed=0;
static int charBuffer[NKEYBUF];
static int nCharBufUsed=0;
static int nMosBufUsed=0;
static FsMouseEventLog mosBuffer[NKEYBUF];
static int exposure=0;



void FsOpenWindow(const FsOpenWindowOption &)
{
	if(NULL!=fsAfterWindowCreationCallBack)
	{
		(*fsAfterWindowCreationCallBack)(fsAfterWindowCreationCallBackParam);
	}
}

void FsCloseWindow(void)
{
}

void FsMaximizeWindow(void)
{
}

void FsGetWindowSize(int &wid,int &hei)
{
	wid=16;
	hei=16;
}

void FsGetWindowPosition(int &x0,int &y0)
{
	x0=0;
	y0=0;
}

void FsSetWindowTitle(const char windowTitle[])
{
	printf("%s\n",windowTitle);
}

void FsPollDevice(void)
{
	static bool busy=false;
	if(true!=busy)
	{
		busy=true;
		if(NULL!=fsPollDeviceHook)
		{
			(*fsPollDeviceHook)(fsPollDeviceHookParam);
		}
		busy=false;
	}
}

void FsPushOnPaintEvent(void)
{
}

void FsSleep(int ms)
{
#ifdef WIN32
	Sleep(ms);
#else
	fd_set set;
	struct timeval wait;
	wait.tv_sec=ms/1000;
	wait.tv_usec=(ms%1000)*1000;
	FD_ZERO(&set);
	select(0,&set,NULL,NULL,&wait);
#endif
}

long long int FsPassedTime(void)
{
	return 10;
}

long long int FsSubSecondTimer(void)
{
	return time(NULL)*1000;
}

int FsInkey(void)
{
	if(nKeyBufUsed>0)
	{
		int i,keyCode;
		keyCode=keyBuffer[0];
		nKeyBufUsed--;
		for(i=0; i<nKeyBufUsed; i++)
		{
			keyBuffer[i]=keyBuffer[i+1];
		}
		return keyCode;
	}
	return 0;
}

int FsInkeyChar(void)
{
	if(nCharBufUsed>0)
	{
		int i,asciiCode;
		asciiCode=charBuffer[0];
		nCharBufUsed--;
		for(i=0; i<nCharBufUsed; i++)
		{
			charBuffer[i]=charBuffer[i+1];
		}
		return asciiCode;
	}
	return 0;
}

void FsPushKey(int fskey)
{
	if(nKeyBufUsed<NKEYBUF)
	{
		keyBuffer[nKeyBufUsed++]=fskey;
	}
}

void FsPushChar(int c)
{
	if(nCharBufUsed<NKEYBUF)
	{
		charBuffer[nCharBufUsed++]=c;
	}
}

int FsGetKeyState(int fsKeyCode)
{
	return 0;
}

int FsCheckWindowExposure(void)
{
	const int ret=exposure;
	exposure=0;
	return ret;
}

void FsGetMouseState(int &lb,int &mb,int &rb,int &mx,int &my)
{
	lb=0;
	mb=0;
	rb=0;
	mx=0;
	my=0;
}

int FsGetMouseEvent(int &lb,int &mb,int &rb,int &mx,int &my)
{
	lb=0;
	mb=0;
	rb=0;
	mx=0;
	my=0;
	return FSMOUSEEVENT_NONE;
}

void FsSwapBuffers(void)
{
}

void FsChangeToProgramDir(void)
{
}

// The following two dummy functions are for loop-holing fsguilib to fssimplewindow for
// managing clipboard.  There's got to be a better way, but I cannot find so far.
void FsX11GetClipBoardString(long long int &,char *&)
{
}
void FsX11SetClipBoardString(long long int,const char [])
{
}
