#ifndef YSTHREAD___IS_INCLUDED
#define YSTHREAD___IS_INCLUDED
/* { */

#include <ysclass.h>
#include <ystask.h>
#include <thread>



/*! This function runs multiple tasks in parallel. */
void YsRunTaskParallel(YSSIZE_T nTask,YsTask *const task[]);



/*! This function runs multiple tasks in parallel. */
template <class T,const int N>
void YsRunTaskParallelDirect(YsArray <T,N> &taskArray)
{
	if(1==taskArray.GetN())
	{
		taskArray[0].StartLocal();
	}
	else if(0<taskArray.GetN())
	{
		const auto nTask=taskArray.GetN();
		YsArray <std::thread,64> threadArray;
		threadArray.Resize(nTask-1);
		for(int idx=0; idx<nTask-1; ++idx)
		{
			std::thread t(YsTask::Start,&taskArray[idx]);
			threadArray[idx]=(std::thread &&)t;
		}
		taskArray[nTask-1].StartLocal();
		for(int idx=0; idx<nTask-1; ++idx)
		{
			threadArray[idx].join();
		}
	}
}



/*! This function runs multiple tasks in parallel.
    T needs to be a class derived from YsTask.  
    Tasks in 'taskArray' may be input and output.  Therefore it shouldn't made const. */
template <const int N>
void YsRunTaskParallel(const YsArray <YsTask *,N> &taskArray)
{
	YsRunTaskParallel(taskArray.GetN(),taskArray);
}



/* } */
#endif
