/* ////////////////////////////////////////////////////////////

Polygon Crest
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

File Name: fsgui3dapp_draw.cpp
//////////////////////////////////////////////////////////// */

#include "ysgebl_gui_editor_base.h"
#include "miscdlg.h"
#include "refbmp/refbmp.h"
#include <ysgl.h>
#include "ysglbuffer.h"
#include "config/config.h"


void GeblGuiEditorBase::DrawGuiOnly(void)
{
	int viewport[4];
	drawEnv.GetOpenGlCompatibleViewport(viewport);
	drawEnv.SetVerticalOrientation(YSTRUE);
	drawEnv.UpdateNearFar();

	SetNeedRedraw(YSFALSE);
	drawEnv.SetNeedRedraw(YSFALSE);
	threeDInterface.SetNeedRedraw(YSFALSE);

	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glDisable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0,(double)viewport[2],(double)viewport[3],0.0,-1.0,1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	FsGuiCanvas::Show();
	FsSwapBuffers();
}

void GeblGuiEditorBase::Draw(void)
{
	if(YSTRUE!=drawingMasterSwitch)
	{
		DrawGuiOnly();
		return;
	}

	// Do this at the beginning of Draw funtion.  This will allow one of the elements set SetNeedRedraw(YSTRUE) 
	// within drawing function so that Draw function will be called again in the next iteragion. >>
	SetNeedRedraw(YSFALSE);
	drawEnv.SetNeedRedraw(YSFALSE);
	threeDInterface.SetNeedRedraw(YSFALSE);
	// <<


	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(1,1);


	if(NULL==slHd)
	{
		if(NULL!=GetTopStatusBar())
		{
			GetTopStatusBar()->ClearStringAll();
		}
		if(NULL!=GetBottomStatusBar())
		{
			GetBottomStatusBar()->ClearStringAll();
		}
	}

	if(NULL!=slHd && NULL!=GetTopStatusBar())
	{
		YsWString ful,pth,fil;
		slHd->GetFileName(ful);
		ful.SeparatePathFile(pth,fil);
		if(YSTRUE==slHd->IsModified())
		{
			fil.Append(L"(*)");
		}

		YsWString curMsg;
		if(0!=YsWString::Strcmp(fil,GetTopStatusBar()->GetString(curMsg,0)))
		{
			GetTopStatusBar()->SetString(0,fil);
		}
	}

	if(NULL!=slHd && NULL!=GetBottomStatusBar())
	{
		YsString str("Selection");
		YsString append;

		{
			YsArray <YsShellVertexHandle> selVtHd;
			slHd->GetSelectedVertex(selVtHd);
			if(0<selVtHd.GetN())
			{
				append.Printf("  Vertex:%d",(int)selVtHd.GetN());
				str.Append(append);
			}
		}
		{
			YsArray <YsShellPolygonHandle> selPlHd;
			slHd->GetSelectedPolygon(selPlHd);
			if(0<selPlHd.GetN())
			{
				append.Printf("  Polygon:%d",(int)selPlHd.GetN());
				str.Append(append);
			}
		}
		{
			YsArray <YsShellExt::ConstEdgeHandle> selCeHd;
			slHd->GetSelectedConstEdge(selCeHd);
			if(0<selCeHd.GetN())
			{
				append.Printf("  ConstEdge:%d",(int)selCeHd.GetN());
				str.Append(append);
			}
		}
		{
			YsArray <YsShellExt::FaceGroupHandle> selFgHd;
			slHd->GetSelectedFaceGroup(selFgHd);
			if(0<selFgHd.GetN())
			{
				append.Printf("  FaceGroup:%d",(int)selFgHd.GetN());
				str.Append(append);
			}
		}
		GetBottomStatusBar()->SetString(0,str);
	}


	int viewport[4];
	drawEnv.GetOpenGlCompatibleViewport(viewport);
	drawEnv.SetVerticalOrientation(YSTRUE);
	drawEnv.UpdateNearFar();


	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);


	if(YsCoordSysModel==YSOPENGL)
	{
		glFrontFace(GL_CCW);
	}
	else
	{
		glFrontFace(GL_CW);
	}


	const YsMatrix4x4 &projMat=drawEnv.GetProjectionMatrix();
	const YsMatrix4x4 &viewMat=drawEnv.GetViewMatrix();
	const YsMatrix4x4 projViewMat=(projMat*viewMat);

	threeDInterface.SetViewport(viewport);
	threeDInterface.SetProjViewModelMatrix(projViewMat);


	glViewport(viewport[0],viewport[1],viewport[2],viewport[3]);

	GLfloat glProjMat[16];
	projMat.GetOpenGlCompatibleMatrix(glProjMat);

	GLfloat glModelviewMat[16];
	viewMat.GetOpenGlCompatibleMatrix(glModelviewMat);


	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMultMatrixf(glProjMat);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMultMatrixf(glModelviewMat);


	YsGLSLSetShared3DRendererProjection(glProjMat);
	YsGLSLSetShared3DRendererModelView(glModelviewMat);


	YsVec3 lightDir=YsUnitVector(YsVec3(0.1,1.0,3.0));
	const GLfloat lightDirf[]={lightDir.xf(),lightDir.yf(),lightDir.zf()};
	YsGLSLSetShared3DRendererDirectionalLightInCameraCoordinatefv(0,lightDirf);
	YsGLSLSetShared3DRendererSpecularColor(config->GetSpecularColor());
	YsGLSLSetShared3DRendererAmbientColor(config->GetAmbientColor());
	YsGLSLSetShared3DRendererLightColor(0,config->GetLightColor());



	actuallyDrawVertex=drawVertex;
	actuallyDrawConstEdge=drawConstEdge;
	actuallyDrawPolygonEdge=drawPolygonEdge;
	actuallyDrawPolygon=drawPolygon;
	actuallyDrawShrunkPolygon=drawShrunkPolygon;
	actuallyDrawNonManifoldEdge=drawNonManifoldEdge;

	if(NULL!=UIBeforeDrawCallBack)
	{
		(*UIBeforeDrawCallBack)(*this);
	}

	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);



	{
// Linear interpolation looks to be too fast.  Probably need three colors.
		YsGLSLUsePlain2DRenderer(YsGLSLSharedPlain2DRenderer());
		YsGLSLUseWindowCoordinateInPlain2DDrawing(YsGLSLSharedPlain2DRenderer(),YSTRUE);

		glDepthMask(0);

		const GLfloat rect[12]=
		{
			0                   ,0,
			(GLfloat)viewport[2],0,
			0                   ,(GLfloat)viewport[3]/2.0f,
			(GLfloat)viewport[2],(GLfloat)viewport[3]/2.0f,
			0                   ,(GLfloat)viewport[3],
			(GLfloat)viewport[2],(GLfloat)viewport[3]
		};
		const GLfloat color[24]=
		{
			config->bgColor[0].Rf(),config->bgColor[0].Gf(),config->bgColor[0].Bf(),1,
			config->bgColor[0].Rf(),config->bgColor[0].Gf(),config->bgColor[0].Bf(),1,
			config->bgColor[1].Rf(),config->bgColor[1].Gf(),config->bgColor[1].Bf(),1,
			config->bgColor[1].Rf(),config->bgColor[1].Gf(),config->bgColor[1].Bf(),1,
			config->bgColor[2].Rf(),config->bgColor[2].Gf(),config->bgColor[2].Bf(),1,
			config->bgColor[2].Rf(),config->bgColor[2].Gf(),config->bgColor[2].Bf(),1
		};
		YsGLSLDrawPlain2DPrimitivefv(YsGLSLSharedPlain2DRenderer(),GL_TRIANGLE_STRIP,6,rect,color);

		glDepthMask(1);

		YsGLSLEndUsePlain2DRenderer(YsGLSLSharedPlain2DRenderer());
	}



	threeDInterface.Draw();


	if(NULL!=slHd && YSTRUE==slHd->NeedRemakeHighlightVertexBuffer())
	{
		slHd->GetHighlight().RemakeVertexBuffer(*slHd);
		slHd->SetNeedRemakeHighlightVertexBuffer(YSFALSE);
	}

	if(NEED_REMAKE_DRAWING_NONE!=needRemakeDrawingBuffer)
	{
		RemakeDrawingBuffer(slHd,needRemakeDrawingBuffer);
		needRemakeDrawingBuffer=NEED_REMAKE_DRAWING_NONE;
	}
	if(NEED_REMAKE_DRAWING_NONE!=needRemakeDrawingBuffer_for_OtherShell && YSTRUE==actuallyDrawOtherShell)
	{
		YsArray <YsShellDnmContainer <YsShellExtEditGui>::Node *> allNode;
		shlGrp.GetNodePointerAll(allNode);
		for(YSSIZE_T nodeIdx=0; nodeIdx<allNode.GetN(); ++nodeIdx)
		{
			if(slHd!=allNode[nodeIdx])
			{
				RemakeDrawingBuffer(allNode[nodeIdx],needRemakeDrawingBuffer_for_OtherShell);
			}
		}
		needRemakeDrawingBuffer_for_OtherShell=NEED_REMAKE_DRAWING_NONE;
	}

	YsShellExtDrawingBuffer &drawBuf=slHd->GetDrawingBuffer();

	if(YSTRUE==actuallyDrawVertex && NULL!=slHd)
	{
		GeblGuiFoundation::DrawVertex(*slHd,viewport);
	}

	if(0<drawBuf.selectedVertexPosBuffer.GetNumVertex() && NULL!=slHd)
	{
		GeblGuiFoundation::DrawSelectedVertex(*slHd,viewport);
	}

	if(0<drawBuf.selectedVertexLineBuffer.GetNumVertex() && NULL!=slHd)
	{
		GeblGuiFoundation::DrawSelectedVertexLine(*slHd);
	}

	if(YSTRUE==actuallyDrawPolygonEdge && NULL!=slHd)
	{
		GeblGuiFoundation::DrawPolygonEdge(*slHd);
	}

	if(YSTRUE==actuallyDrawShrunkPolygon && NULL!=slHd)
	{
		GeblGuiFoundation::DrawShrunkPolygon(*slHd);

	}

	if(0<drawBuf.selectedPolygonPosBuffer.GetNumVertex() && NULL!=slHd)
	{
		GeblGuiFoundation::DrawSelectedPolygon(*slHd);
	}

	if(YSTRUE==actuallyDrawConstEdge && 0<drawBuf.constEdgeVtxBuffer.GetNumVertex() && NULL!=slHd)
	{
		GeblGuiFoundation::DrawConstEdge(*slHd);
	}

	if(0<drawBuf.selectedConstEdgeVtxBuffer.GetNumVertex() && NULL!=slHd)
	{
		GeblGuiFoundation::DrawSelectedConstEdge(*slHd);
	}

	if(YSTRUE!=dnmEditMode)
	{
		DrawInGeometryEditMode();
	}
	else
	{
		DrawInDnmEditMode();
	}

	if(0<=drawBuf.selectedFaceGroupVtxBuffer.GetNumVertex() && NULL!=slHd)
	{
		GeblGuiFoundation::DrawSelectedFaceGroup(*slHd);
	}


	if(NULL!=slHd && YSTRUE==actuallyDrawNonManifoldEdge)
	{
		if(YSTRUE==nonManifoldEdgeCache.IsModified(slHd))
		{
			nonManifoldEdgeCache.CacheNonManifoldEdge(slHd);
		}
		struct YsGLSL3DRenderer *renderer=YsGLSLSharedVariColor3DRenderer();
		YsGLSLUse3DRenderer(renderer);

		glLineWidth(3);
		YsGLSLDrawPrimitiveVtxColfv(renderer,GL_LINES,nonManifoldEdgeCache.lineVtxBuf.GetNumVertex(),nonManifoldEdgeCache.lineVtxBuf,nonManifoldEdgeCache.lineColBuf);
		glLineWidth(1);

		YsGLSLEndUse3DRenderer(renderer);

		// The following could have easily be done by glPointSize.  However, stupid OpenGL ES 2.0 got rid of glPointSize.
		// OpenGL ES should be gone.  The hardware on the tablets and smart phones are good enough to support full-blown
		// OpenGL.  OpenGL ES will be remembered as a dark history of OpenGL.  That is bad.  Not as bad as Direct 3D though.
		renderer=YsGLSLSharedVariColorMarker3DRenderer();
		YsGLSLUse3DRenderer(renderer);

		YsGLSLSet3DRendererViewportDimensionf(renderer,(GLfloat)viewport[2],(GLfloat)viewport[3]);
		YsGLSLSet3DRendererUniformMarkerType(renderer,YSGLSL_MARKER_TYPE_RECT);
		YsGLSLSet3DRendererUniformMarkerDimension(renderer,1.5);
		YsGLSLDrawPrimitiveVtxColPixelOffsetfv(
		    renderer,GL_TRIANGLES,
		    nonManifoldEdgeCache.pointVtxBuf.GetNumVertex(),
		    nonManifoldEdgeCache.pointVtxBuf,
		    nonManifoldEdgeCache.pointColBuf,
		    nonManifoldEdgeCache.pointPixOffsetBuf);
		YsGLSLEndUse3DRenderer(renderer);
	}

	if(YSTRUE==drawHighlight)
	{
		auto renderer=YsGLSLSharedFlat3DRenderer();
		YsGLSLUse3DRenderer(renderer);

		const GLfloat blue[]={0.0f,0.0f,1.0f,1.0f};
		YsGLSLSet3DRendererUniformColorfv(renderer,blue);
		YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,slHd->GetHighlight().highlightedEdgeVertexBuffer.GetNumVertex(),slHd->GetHighlight().highlightedEdgeVertexBuffer);

		const GLfloat darkBlue[]={0.0f,0.0f,0.7f,1.0f};
		YsGLSLSet3DRendererUniformColorfv(renderer,darkBlue);
		YsGLSLDrawPrimitiveVtxfv(renderer,GL_TRIANGLES,slHd->GetHighlight().highlightedPolygonVertexBuffer.GetNumVertex(),slHd->GetHighlight().highlightedPolygonVertexBuffer);

		YsGLSLEndUse3DRenderer(renderer);
	}

	if(NULL!=UIDrawCallBack3D)
	{
		UIDrawCallBack3D(*this);
	}


	refBmpStore->Draw();


	if(YSTRUE==drawAxis)
	{
		const GLfloat axisLength=5.0;
		GeblGuiFoundation::DrawAxis(axisLength,viewport,projMat,viewMat);
	}



	// 2D Drawing
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0,(double)viewport[2],(double)viewport[3],0.0,-1.0,1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	YsGLSLUsePlain2DRenderer(YsGLSLSharedPlain2DRenderer());
	YsGLSLUseWindowCoordinateInPlain2DDrawing(YsGLSLSharedPlain2DRenderer(),YSTRUE);
	{
		if(YSTRUE==lBtnDown)
		{
			const GLfloat rect[8]=
			{
				(GLfloat)lBtnDownMx,(GLfloat)lBtnDownMy,
				(GLfloat)moveMx,    (GLfloat)lBtnDownMy,
				(GLfloat)moveMx,    (GLfloat)moveMy,
				(GLfloat)lBtnDownMx,(GLfloat)moveMy
			};
			const GLfloat color[16]=
			{
				0,0,0,0.8f,
				0,0,0,0.8f,
				0,0,0,0.8f,
				0,0,0,0.8f
			};
			YsGLSLDrawPlain2DPrimitivefv(YsGLSLSharedPlain2DRenderer(),GL_LINE_LOOP,4,rect,color);
		}
	}
	YsGLSLEndUsePlain2DRenderer(YsGLSLSharedPlain2DRenderer());

	glDisable(GL_DEPTH_TEST);

	FsGuiCanvas::Show();

	if(NULL!=UIDrawCallBack2D)
	{
		(*UIDrawCallBack2D)(*this);
	}

	FsSwapBuffers();
}

void GeblGuiEditorBase::DrawInGeometryEditMode(void)
{
	if(YSTRUE==actuallyDrawPolygon && NULL!=slHd)
	{
		YsShellExtDrawingBuffer &buf=slHd->GetDrawingBuffer();

		glEnable(GL_CULL_FACE);

		if(NULL!=YsGLSLSharedVariColorShaded3DRenderer())
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

			// 2014/12/11 Looks like there may be too many triangles.
			const YSSIZE_T perPrim=99999; // Needs to be 3*N
			for(YSSIZE_T offset=0; offset<buf.solidShadedVtxBuffer.GetNumVertex(); offset+=perPrim)
			{
				YSSIZE_T nVtx=YsSmaller(buf.solidShadedVtxBuffer.GetNumVertex()-offset,perPrim);

				YsGLSLDrawPrimitiveVtxNomColfv(
				    YsGLSLSharedVariColorShaded3DRenderer(),
				    GL_TRIANGLES,
				    (int)nVtx,
				    buf.solidShadedVtxBuffer+3*offset,
				    buf.solidShadedNomBuffer+3*offset,
				    buf.solidShadedColBuffer+4*offset);
			}
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		}
		if(NULL!=YsGLSLSharedVariColor3DRenderer())
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());

			YsGLSLDrawPrimitiveVtxColfv(
			    YsGLSLSharedVariColor3DRenderer(),
			    GL_TRIANGLES,
			    buf.solidUnshadedVtxBuffer.GetNumVertex(),
			    buf.solidUnshadedVtxBuffer,
			    buf.solidUnshadedColBuffer);

			YsGLSLDrawPrimitiveVtxColfv(
			    YsGLSLSharedVariColor3DRenderer(),
			    GL_TRIANGLES,
			    buf.backFaceVtxBuffer.GetNumVertex(),
			    buf.backFaceVtxBuffer,
			    buf.backFaceColBuffer);

			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		}
		if(NULL!=YsGLSLSharedVariColorShaded3DRenderer())
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
			YsGLSLDrawPrimitiveVtxNomColfv(
			    YsGLSLSharedVariColorShaded3DRenderer(),
			    GL_TRIANGLES,
			    buf.trspShadedVtxBuffer.GetNumVertex(),
			    buf.trspShadedVtxBuffer,
			    buf.trspShadedNomBuffer,
			    buf.trspShadedColBuffer);
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		}
		if(NULL!=YsGLSLSharedVariColor3DRenderer())
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
			YsGLSLDrawPrimitiveVtxColfv(
			    YsGLSLSharedVariColor3DRenderer(),
			    GL_TRIANGLES,
			    buf.trspUnshadedVtxBuffer.GetNumVertex(),
			    buf.trspUnshadedVtxBuffer,
			    buf.trspUnshadedColBuffer);
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		}
		
		glDisable(GL_CULL_FACE);
	}

	if(YSTRUE==actuallyDrawOtherShell)
	{
		YsArray <const YsShellDnmContainer <YsShellExtEditGui>::Node *,16> allNode;
		shlGrp.GetNodePointerAll(allNode);
		for(YSSIZE_T nodeIdx=0; nodeIdx<allNode.GetN(); ++nodeIdx)
		{
			if(slHd!=allNode[nodeIdx])
			{
				struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
				YsShellExtDrawingBuffer &buf=allNode[nodeIdx]->GetDrawingBuffer();
				YsGLSLUse3DRenderer(renderer);

				const GLfloat edgeCol[4]={0,0,0,0.2f};
				YsGLSLSet3DRendererUniformColorfv(renderer,edgeCol);
				YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,buf.normalEdgePosBuffer.GetNumVertex(),buf.normalEdgePosBuffer);

				YsGLSLEndUse3DRenderer(renderer);
			}
		}
	}
}

void GeblGuiEditorBase::DrawInDnmEditMode(void)
{
	auto renderer=YsGLSLSharedVariColorShaded3DRenderer();
	GLfloat curGLTfm[16];

	YsGLSLUse3DRenderer(renderer);
	YsGLSLGet3DRendererModelViewfv(curGLTfm,renderer);
	YsGLSLEndUse3DRenderer(renderer);

	YsMatrix4x4 curTfm(YSFALSE);
	curTfm.CreateFromOpenGlCompatibleMatrix(curGLTfm);

	glEnable(GL_CULL_FACE);

	auto rootArray=shlGrp.GetRootNodeArray();
	for(auto node : rootArray)
	{
		DrawInDnmEditModeRecursive(node,dnmState,curTfm);
	}

	YsGLSL3DRenderer *usedRenderer[4]=
	{
		YsGLSLSharedVariColorShaded3DRenderer(),
		YsGLSLSharedVariColor3DRenderer(),
		YsGLSLSharedVariColorShaded3DRenderer(),
		YsGLSLSharedVariColor3DRenderer(),
	};
	for(int i=0; i<4; ++i)
	{
		YsGLSLUse3DRenderer(usedRenderer[i]);
		YsGLSLSet3DRendererModelViewfv(usedRenderer[i],curGLTfm);
		YsGLSLEndUse3DRenderer(usedRenderer[i]);
	}

	glDisable(GL_CULL_FACE);
}

void GeblGuiEditorBase::DrawInDnmEditModeRecursive(YsShellDnmContainer <YsShellExtEditGui>::Node *node,const YsShellDnmContainer <YsShellExtEditGui>::DnmState &state,const YsMatrix4x4 &curTfm)
{
	const YsMatrix4x4 tfm=curTfm*state.GetTransformation(node);

	GLfloat glTfm[16];
	tfm.GetOpenGlCompatibleMatrix(glTfm);

	for(auto childNode : node->children)
	{
		DrawInDnmEditModeRecursive(childNode,state,tfm);
	}

	if(YSTRUE==state.GetShow(node))
	{
		auto &buf=node->GetDrawingBuffer();

		if(NULL!=YsGLSLSharedVariColorShaded3DRenderer())
		{
			auto renderer=YsGLSLSharedVariColorShaded3DRenderer();
			YsGLSLUse3DRenderer(renderer);

			YsGLSLSet3DRendererModelViewfv(renderer,glTfm);
			YsGLSLDrawPrimitiveVtxNomColfv(
			    renderer,
			    GL_TRIANGLES,
			    buf.solidShadedVtxBuffer.GetNumVertex(),
			    buf.solidShadedVtxBuffer,
			    buf.solidShadedNomBuffer,
			    buf.solidShadedColBuffer);
			YsGLSLEndUse3DRenderer(renderer);
		}
		if(NULL!=YsGLSLSharedVariColor3DRenderer())
		{
			auto renderer=YsGLSLSharedVariColor3DRenderer();
			YsGLSLUse3DRenderer(renderer);

			YsGLSLSet3DRendererModelViewfv(renderer,glTfm);
			YsGLSLDrawPrimitiveVtxColfv(
			    renderer,
			    GL_TRIANGLES,
			    buf.solidUnshadedVtxBuffer.GetNumVertex(),
			    buf.solidUnshadedVtxBuffer,
			    buf.solidUnshadedColBuffer);

			YsGLSLEndUse3DRenderer(renderer);
		}
		if(NULL!=YsGLSLSharedVariColorShaded3DRenderer())
		{
			auto renderer=YsGLSLSharedVariColorShaded3DRenderer();
			YsGLSLUse3DRenderer(renderer);

			YsGLSLSet3DRendererModelViewfv(renderer,glTfm);
			YsGLSLDrawPrimitiveVtxNomColfv(
			    renderer,
			    GL_TRIANGLES,
			    buf.trspShadedVtxBuffer.GetNumVertex(),
			    buf.trspShadedVtxBuffer,
			    buf.trspShadedNomBuffer,
			    buf.trspShadedColBuffer);
			YsGLSLEndUse3DRenderer(renderer);
		}
		if(NULL!=YsGLSLSharedVariColor3DRenderer())
		{
			auto renderer=YsGLSLSharedVariColor3DRenderer();
			YsGLSLUse3DRenderer(renderer);

			YsGLSLSet3DRendererModelViewfv(renderer,glTfm);
			YsGLSLDrawPrimitiveVtxColfv(
			    renderer,
			    GL_TRIANGLES,
			    buf.trspUnshadedVtxBuffer.GetNumVertex(),
			    buf.trspUnshadedVtxBuffer,
			    buf.trspUnshadedColBuffer);
			YsGLSLEndUse3DRenderer(renderer);
		}
	}
}

void GeblGuiEditorBase::RemakeDrawingBuffer(
    YsShellDnmContainer <YsShellExtEditGui>::Node *slHd,
    unsigned int needRemakeDrawingBuffer)
{
	YsShellExtDrawingBuffer &buf=slHd->GetDrawingBuffer();

	buf.drawBackFaceInDifferentColor=drawBackFaceInDifferentColor;
	buf.backFaceColor=backFaceColor;

	if(0!=(needRemakeDrawingBuffer & NEED_REMAKE_DRAWING_VERTEX) && NULL!=slHd)
	{
		buf.RemakeNormalVertexBuffer(slHd->Conv(),normalVertexMarkerRadius,roundVertexMarkerRadius);
	}

	if(0!=(needRemakeDrawingBuffer & NEED_REMAKE_DRAWING_SELECTED_VERTEX))
	{
		YsArray <YsShellVertexHandle> selVtHd;
		slHd->GetSelectedVertex(selVtHd);
		buf.RemakeSelectedVertexBuffer(slHd->Conv(),selectedVertexMarkerRadius,selVtHd);
	}

	if(0!=(needRemakeDrawingBuffer & NEED_REMAKE_DRAWING_POLYGON))
	{
		buf.RemakePolygonBuffer(slHd->Conv(),polygonShrinkRatioForPicking);
	}

	if(0!=(needRemakeDrawingBuffer & NEED_REMAKE_DRAWING_SELECTED_POLYGON))
	{
		YsArray <YsShellPolygonHandle> selPlHd;
		slHd->GetSelectedPolygon(selPlHd);
		buf.RemakeSelectedPolygonBuffer(slHd->Conv(),polygonShrinkRatioForPicking,selPlHd);
	}

	if(0!=(needRemakeDrawingBuffer & (NEED_REMAKE_DRAWING_CONSTEDGE|NEED_REMAKE_DRAWING_SELECTED_CONSTEDGE)))
	{
		YsArray <YsShellExt::ConstEdgeHandle> selCeHd;
		slHd->GetSelectedConstEdge(selCeHd);
		buf.RemakeConstEdgeBuffer(slHd->Conv(),selCeHd);
		buf.RemakeSelectedConstEdgeBuffer(slHd->Conv(),selCeHd);
	}

	if(0!=(needRemakeDrawingBuffer & (NEED_REMAKE_DRAWING_SELECTED_FACEGROUP)))
	{
		YsArray <YsShellExt::FaceGroupHandle> selFgHd;
		slHd->GetSelectedFaceGroup(selFgHd);
		buf.RemakeSelectedFaceGroupBuffer(slHd->Conv(),selFgHd);
	}
}

