/* ////////////////////////////////////////////////////////////

Polygon Crest
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

File Name: fsgui3dmain.cpp
//////////////////////////////////////////////////////////// */

#include <fssimplewindow.h>
#include <ysdebug.h>
#include <ysport.h>

// Comment out FSGUI3D_USE_SYSTEM_FONT to remove depencency on YsSystemFont library.
#define FSGUI3D_USE_SYSTEM_FONT

// Comment out FSGUI3D_USE_MODERN_UI to use primitive graphics, which may be faster.
#define FSGUI3D_USE_MODERN_UI



#ifdef FSGUI3D_USE_SYSTEM_FONT
#include <ysfontrenderer.h>
#endif

#include "ysgebl_gui_editor_base.h"
#include "filename.h"
#include "config/config.h"
#include "ysgebl_gui_extension_base.h"

// Modify this program if the initial window location needs to be changed.


static void UpdateWindowPosition(int &x0,int &y0,int &wid,int &hei)
{
	auto *cfg=PolyCreConfig::Create();
	auto *fileAndDirName=PolyCreFileName::Create();

	cfg->Load(fileAndDirName->GetOptionFileName());
	if(YSTRUE==cfg->saveWindowPositionAndSize)
	{
		auto wposFn=fileAndDirName->GetLastWindowPositionFileName();

		FILE *fp=YsFileIO::Fopen(wposFn,"r");
		if(NULL!=fp)
		{
			YsString str;
			while(NULL!=str.Fgets(fp))
			{
				YsArray <YsString> args;
				str.Arguments(args);
				if(0==strcmp(args[0],"POS") && 5<=args.GetN())
				{
					x0=atoi(args[1]);
					y0=atoi(args[2]);
					wid=atoi(args[3]);
					hei=atoi(args[4]);
					if(0>x0)
					{
						x0=0;
					}
					if(0>y0)
					{
						y0=0;
					}
					if(120>wid)
					{
						wid=120;
					}
					if(120>hei)
					{
						hei=120;
					}
				}
			}
			fclose(fp);
		}
	}

	PolyCreConfig::Delete(cfg);
	PolyCreFileName::Delete(fileAndDirName);
}


int main(int ac,char *av[])
{
	int x0=96,y0=0,wid=1200,hei=640;
	UpdateWindowPosition(x0,y0,wid,hei);
	FsOpenWindow(x0,y0,wid,hei,1);


#ifdef FSGUI3D_USE_SYSTEM_FONT
	YsSystemFontRenderer *sysFont=new YsSystemFontRenderer;
	FsGuiObject::defUnicodeRenderer=sysFont;
#endif

#ifdef FSGUI3D_USE_MODERN_UI
	FsGuiObject::scheme=FsGuiObject::MODERN;
	FsGuiObject::defRoundRadius=8.0;
	FsGuiObject::defHScrollBar=20;
	FsGuiObject::defHAnnotation=14;
	FsGuiObject::defVSpaceUnit=12;

	FsGuiObject::defDialogBgCol.SetDoubleRGB(0.75,0.75,0.75);

	FsGuiObject::defTabBgCol.SetDoubleRGB(0.82,0.82,0.82);
	FsGuiObject::defTabClosedFgCol.SetDoubleRGB(0.8,0.8,0.8);
	FsGuiObject::defTabClosedBgCol.SetDoubleRGB(0.2,0.2,0.2);

	FsGuiObject::defBgCol.SetDoubleRGB(0.85,0.85,0.85);
	FsGuiObject::defFgCol.SetDoubleRGB(0.0,0.0,0.0);
	FsGuiObject::defActiveBgCol.SetDoubleRGB(0.3,0.3,0.7);
	FsGuiObject::defActiveFgCol.SetDoubleRGB(1.0,1.0,1.0);
	FsGuiObject::defFrameCol.SetDoubleRGB(0.0,0.0,0.0);

	FsGuiObject::defListFgCol.SetDoubleRGB(0.0,0.0,0.0);
	FsGuiObject::defListBgCol.SetDoubleRGB(0.8,0.8,0.8);
	FsGuiObject::defListActiveFgCol.SetDoubleRGB(1.0,1.0,1.0);
	FsGuiObject::defListActiveBgCol.SetDoubleRGB(0.3,0.3,0.7);
#endif

	int customFontHeight=PolyCreGetCustomFontHeight();
	if(0==customFontHeight)
	{
		customFontHeight=16;
	}

	FsGuiObject::defUnicodeRenderer->RequestDefaultFontWithPixelHeight(customFontHeight);
	FsGuiObject::defAsciiRenderer->RequestDefaultFontWithPixelHeight(customFontHeight);

	YsCoordSysModel=YSOPENGL;

	FsGui3DAppInitialize(ac,av);

	while(YSTRUE!=FsGui3dAppMustTerminate())
	{
		FsPollDevice();
		FsGui3DAppIntervalCallBack();
		FsSleep(FsGui3dAppSleepIntervalInMS());
	}

	FsGui3dAppCleanUp();

#ifdef FSGUI3D_USE_SYSTEM_FONT
	delete sysFont;
#endif

	FsSleep(50); // Windows apparently has a problem terminating within certain time after last rendering.  Wait 50 ms.

	exit(0); // This typically is quicker, otherwise destructors occasinally takes long time to finish.
	return 0;
}

#ifdef _WIN32
// Infamouse WinMain function.  This function will be used when SUBSYSTEM is set to WINDOWS.
#include <windows.h>
int PASCAL WinMain(HINSTANCE,HINSTANCE,LPSTR param,int)
{
	int argc=((NULL==param || param[0]==0) ? 1 : 2);
	char *argv[2]=
	{
		"dummy.exe",
		(char *)param
	};
	return main(argc,argv);
}
#endif
