/* ////////////////////////////////////////////////////////////

Polygon Crest
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

File Name: ysgebl_gui_editor_base.h
//////////////////////////////////////////////////////////// */

#ifndef FSGUI_DAPP_IS_INCLUDED
#define FSGUI_DAPP_IS_INCLUDED
/* { */

#include <thread>
#include <mutex>

#include <fsgui3d.h>

#include <ysshellkernel.h>
#include <ysglbuffer.h>

#include <ysgebl_gui_foundation.h>

#include <ysshellextedit_duplicateutil.h>

// Call back functions from the main framework. >>
YSBOOL FsGui3dAppMustTerminate(void);
int FsGui3dAppSleepIntervalInMS(void);

void FsGui3DAppInitialize(int ac,char *av[]);
void FsGui3DAppIntervalCallBack(void);
void FsGui3dAppCleanUp(void);
// Call back functions from the main framework. <<

class GeblGuiEditorMainMenu : public FsGuiPopUpMenu
{
private:
	class GeblGuiEditorBase *canvas;
public:
	GeblGuiEditorMainMenu(class GeblGuiEditorBase *canvas);
	virtual void OnSelectMenuItem(FsGuiPopUpMenuItem *item);

	class GeblGuiEditorBase *GetMainCanvas(void) const;
};

class GeblGuiEditorBase : public GeblGuiFoundation
{
friend class PolyCreProjectionDialog;
friend class PolyCreEditSweepConstEdgeDialog;
friend class PolyCreSlideShowDialog;
friend class PolyCreDnmTreeDialog;
friend class GeblGuiExtension;

public:
	typedef GeblGuiEditorBase THISCLASS;

	class ShortCutKey
	{
	public:
		void (*callBackFunc)(void *,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
		int fskey;
		YSBOOL ctrl,shift;
	};

	enum
	{
		MENU_FILE_EXIT
	};



	FsGuiMenuCallBackConnector <THISCLASS> menuConn;
	int menuIdSource;
	FsGuiPopUpMenuItem *AddMenuItem(FsGuiPopUpMenu *parent,int fsKey,const char label[],void (THISCLASS::*funcPtr)(FsGuiPopUpMenuItem *));
	FsGuiPopUpMenuItem *AddMenuItem(FsGuiPopUpMenu *parent,int fsKey,const wchar_t label[],void (THISCLASS::*funcPtr)(FsGuiPopUpMenuItem *));

	FsGuiDialogCallBackFunction <THISCLASS,int> closeModalCallBack;
	void RegisterCloseModalCallBack(void (THISCLASS::*funcPtr)(FsGuiDialog *,int));


	std::mutex guiLock;
	YSBOOL drawingMasterSwitch; // If false, nothing other than gui widgets will be drawn.



	YsWString lastAccessedFileName;

	class PolyCreFileName *fileAndDirName;
	class PolyCreConfig *config;
	class PolyCreReferenceBitmapStore *refBmpStore;

	FsGui3DInterface threeDInterface;
	FsGui3DViewControlDialogBase *viewControlDlg;
	GeblGuiEditorMainMenu *mainMenu;
	FsGuiStatusBar *topBar,*bottomBar;
	class PolyCreSketchInterface *sketchUI;

	FsGuiPopUpMenu *fileRecent;
	FsGuiRecentFiles recent;

	FsGuiPopUpMenuItem *viewDrawAxis;
	FsGuiPopUpMenuItem *viewDrawVertex;
	FsGuiPopUpMenuItem *viewDrawConstEdge;
	FsGuiPopUpMenuItem *viewDrawPolygonFill;
	FsGuiPopUpMenuItem *viewDrawWireframe;
	FsGuiPopUpMenuItem *viewDrawShrunkPolygon;
	FsGuiPopUpMenuItem *viewDrawBackFaceInDifferentColor;
	FsGuiPopUpMenuItem *viewEnableViewPort;
	FsGuiPopUpMenuItem *viewEnableCrossSection;
	FsGuiPopUpMenuItem *viewUseSameViewPortForAll;
	FsGuiPopUpMenuItem *viewShowHighlight;
	FsGuiPopUpMenuItem *viewHighlightNonManifoldEdge;

	FsGuiPopUpMenuItem *globalTargetCurrentShellOnly;
	FsGuiPopUpMenuItem *globalTargetAllShell;

	YsShellExtEdit_DuplicateUtil cutBuffer;


	// [Modeless dialogs]
	// Adding a modeless dialog:
	//   (1) Add a pointer in the following chunk.
	//   (2) Add an initialization in the constructor of the aplication.
	//   (3) Add deletion in the destructor of the application.
	//   (4) Add RemoveDialog in Edit_ClearUIIMode
	class PolyCreCrossSectionDialog *crossSectionDlg;
	class PolyCreSetFaceGroupNameDialog *faceNameDlg;
	class PolyCreSolidOfRevolutionDialog *solidOfRevolutionDlg;
	class PolyCreDnmTreeDialog *dnmTreeDlg;
	class PolyCreAngleSliderDialog *angleSliderDlg;
	class PolyCreEquilateralDialog *equilateralDlg;
	class PolyCreSlideShowDialog *slideShowDlg;
	class PolyCreParallelogramDialog *parallelogramDlg;
	class PolyCreInsertVertexDialog *insertVtxDlg;
	class PolyCreChopOffDialog *chopOffDlg;
	class PolyCreExtendOrShrinkDialog *extendOrShrinkDlg;
	class YsColorPaletteDialog *colorPaletteDlg;
	class FsGuiFileDialog *fdlg;
	FsGuiMessageBoxDialog *mbox;
	FsGuiNumberDialog *numberDlg;
	class YsCreateAirfoilDialog *createAirfoilDlg;
	class PolyCreConfigDialog *configDlg;
	class PolyCreRefBmpDialog *refBmpDlg;
	class PolyCreEditParallelSweepDialog *parallelSweepDlg;
	class PolyCreEditOffsetThickeningDialog *thickeningDlg;
	class PolyCreEditParallelSweepWithPathDialog *parallelSweepWithPathDlg;
	class PolyCreEditSweepConstEdgeDialog *sweepConstEdgeDlg;
	class PolyCreEditCreateMirrorImageDialog *mirrorImageDlg;
	class PolyCreHemmingDialog *hemmingDlg;
	class PolyCreRoundDialog *roundDlg;
	class PolyCreCreatePrimitiveDialog *createPrimitiveDlg;
	class PolyCreProjectionDialog *projectionDlg;
	class PolyCreBoolDialog *boolDlg;

	// [Modal dialogs]
	class PolyCreAboutDialog *aboutDlg;
	class PolyCreHelpDialog *helpDlg;


	class YsShellExt_SweepInfo *sweepInfo;
	class YsShellExt_ThickeningInfo *thickeningInfo;

	YSBOOLEANOPERATION boolOpType;

	YSBOOL undoRedoRequireToClearUIMode;
	YSBOOL deletionRequireToClearUIMode;

	YsArray <ShortCutKey> shortCut;

	class GeblGuiExtensionBase *guiExtension;


	GeblGuiEditorBase();
	~GeblGuiEditorBase();
private:
	class DrawingCacheState
	{
	public:
		const YsShellDnmContainer <YsShellExtEditGui>::Node *slHd;
		YsShellExtEdit::SavePoint whenCached;
		void Save(const YsShellDnmContainer <YsShellExtEditGui>::Node *shl);
		YSBOOL IsModified(const YsShellDnmContainer <YsShellExtEditGui>::Node *shl) const;
	};
	class NonManifoldEdgeCache : public DrawingCacheState
	{
	public:
		YsGLVertexBuffer lineVtxBuf;
		YsGLColorBuffer lineColBuf;

		YsGLVertexBuffer pointVtxBuf; // Stupid OpenGL ES2.0 got rid of glPointSize.  All points need to be drawn as triangles.  What a waste.
		YsGLPixOffsetBuffer pointPixOffsetBuf;
		YsGLColorBuffer pointColBuf;

		void CacheNonManifoldEdge(const YsShellDnmContainer <YsShellExtEditGui>::Node *shl);
	};

	void AddViewControlDialog(void);
	void MakeMainMenu(void);
public:
	void SetMenuCheck(void);
private:
	void DeleteMainMenu(void);

public:
	void MessageDialog(const wchar_t title[],const wchar_t msg[],YSBOOL okBtnBottom=YSFALSE);
	void MessageDialog(const wchar_t title[],const wchar_t msg[],const wchar_t copyBtn[]);
	void YesNoDialog(
	    const wchar_t title[],const wchar_t msg[],const wchar_t yes[],const wchar_t no[],
	    void (*closeModalCallBackFunc)(FsGuiControlBase *,FsGuiDialog *,int));

	virtual void OnInterval(void);
	virtual void OnSelectMenuItem(FsGuiPopUpMenuItem *menuItem);
	virtual YSRESULT OnLButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	virtual YSRESULT OnLButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	virtual YSRESULT OnRButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	virtual YSRESULT OnRButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	virtual YSRESULT OnMouseMove(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my,int prevMx,int prevMy);

	virtual YSRESULT OnSaveView(const class FsGui3DViewControlDialog *dlg);
	static void OnSaveView_FileSelected(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	static void OnSaveView_ConfirmOverwriteCallBack(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	static void OnSaveView_OverwriteChecked(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);

	virtual YSRESULT OnLoadView(class FsGui3DViewControlDialog *dlg);
	static void OnLoadView_FileSelected(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);

	void OnModalDialogClosed(int dialogIdent,class FsGuiDialog *closedModalDialog,int modalDialogEndCode);


	YSRESULT (*ModeCleanUpCallBack)(GeblGuiEditorBase &canvas);
	YSRESULT (*SpaceKeyCallBack)(GeblGuiEditorBase &canvas);
	YSRESULT (*BackSpaceKeyCallBack)(GeblGuiEditorBase &canvas);
	YSRESULT (*LButtonCallBack)(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT (*MouseMoveCallBack)(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSBOOL lBtnDown;  // Shouldn't be turned on if lButtonDown is caught by other UI widgets.
	int lBtnDownMx,lBtnDownMy;
	int moveMx,moveMy;
	YSRESULT (*RButtonCallBack)(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT (*LButtonBoxCallBack)(GeblGuiEditorBase &canvas,int xMin,int yMin,int xMax,int yMax);

	YSRESULT OnKeyDown(int fskey,YSBOOL shift,YSBOOL ctrl,YSBOOL alt);
	void DrawGuiOnly(void);
	void Draw(void);
private:
	void DrawInGeometryEditMode(void);
	void DrawInDnmEditMode(void);
	void DrawInDnmEditModeRecursive(YsShellDnmContainer <YsShellExtEditGui>::Node *,const YsShellDnmContainer <YsShellExtEditGui>::DnmState &state,const YsMatrix4x4 &curTfm);

public:
	void RemakeDrawingBuffer(
	    YsShellDnmContainer <YsShellExtEditGui>::Node *slHd,
	    unsigned int needRemakeDrawingBuffer);
	void AddPolygonFragment(YsShellPolygonHandle plHd,const YsVec3 &nom,const YsShellVertexHandle triVtHd[3],YSBOOL flatShaded);

	// PickedSomething and BoxedSomething are moved to gui_foundation.

public:
	enum
	{
		NEED_REMAKE_DRAWING_NONE=0,
		NEED_REMAKE_DRAWING_ALL=~(unsigned int)0,
		NEED_REMAKE_DRAWING_VERTEX=1,
		NEED_REMAKE_DRAWING_SELECTED_VERTEX=2,
		NEED_REMAKE_DRAWING_POLYGON=4,
		NEED_REMAKE_DRAWING_SELECTED_POLYGON=8,
		NEED_REMAKE_DRAWING_CONSTEDGE=16,
		NEED_REMAKE_DRAWING_SELECTED_CONSTEDGE=32,
		NEED_REMAKE_DRAWING_SELECTED_FACEGROUP=64,
	};
	unsigned int needRemakeDrawingBuffer;
	unsigned int needRemakeDrawingBuffer_for_OtherShell;
	YSBOOL drawAxis;
	YSBOOL drawVertex;
	YSBOOL drawPolygonEdge;
	YSBOOL drawPolygon;
	YSBOOL drawConstEdge;
	YSBOOL drawShrunkPolygon;
	YSBOOL drawOtherShell;
	YSBOOL drawBackFaceInDifferentColor;
	YSBOOL drawApplyViewPortToAllShell;
	YSBOOL drawHighlight;
	YsColor backFaceColor;

	int dnmCurrentState; // -1:Static mode
	YSBOOL dnmEditMode;
	YsShellDnmContainer <YsShellExtEditGui>::DnmState dnmState;

private:
	YSBOOL drawNonManifoldEdge,actuallyDrawNonManifoldEdge;
	mutable NonManifoldEdgeCache nonManifoldEdgeCache;

	void (*UIBeforeDrawCallBack)(GeblGuiEditorBase &canvas);
	void (*UIDrawCallBack2D)(GeblGuiEditorBase &canvas);
	void (*UIDrawCallBack3D)(GeblGuiEditorBase &canvas);
	YsGLVertexBuffer UILineVtxBuf;

private:

public:
	void MakeInitialEmptyShell(void);
	void LoadGeneral(const wchar_t fn[],YsShellDnmContainer <YsShellExtEditGui>::Node *parent,YSBOOL updateRecentFiles);
	void SaveGeneral(const YsShellExtEdit &shl,const wchar_t fn[],YSBOOL updateRecentFiles);
	YsWString GetDataDir(void) const;
	void AddShortCut(int fskey,YSBOOL shift,YSBOOL ctrl,void (*callBackFunc)(void *,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *));


	void RefreshRecentlyUsedFileList(void);
	void AddRecentlyUsedFile(const wchar_t wfn[]);


	// Menu Call Back Functions >>
	static void File_New(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void File_Open(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void File_Open_FileSelected(FsGuiControlBase *canvas,FsGuiDialog *,int);
	static void File_OpenDnm_ConfirmCloseCallBack(FsGuiControlBase *canvas,FsGuiDialog *,int);
	void File_Open_AfterOpenFile(YsShellDnmContainer <YsShellExtEditGui>::Node *newSlHd,const wchar_t fn[],YSBOOL recordRecentFiles);

	static void File_Save(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void File_SaveAs(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void File_SaveAs_FileSelected(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	static void File_SaveAs_STLChecked(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	void File_SaveAs_CheckOverwrite(void);
	static void File_SaveAs_ConfirmOverwriteCallBack(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	void File_SaveAs_PassedOverWriteCheck(const wchar_t fn[]);

	static void File_Close(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_Close_ConfirmCloseCallBack(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	void File_Close_Confirmed(void);

	static void File_Exit(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_Exit_ConfirmExitCallBack(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);

	static void File_SlideShow(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_SlideShow_FileSelected(FsGuiControlBase *canvas,FsGuiDialog *,int);
	static YSRESULT File_SlideShow_ModeCleanUpCallBack(GeblGuiEditorBase &canvas);

	static void File_GetInfo_Statistics(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_GetInfo_Volume(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_GetInfo_Length_BetweenVertex(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_GetInfo_Length_ConstEdge(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_GetInfo_Area_SelectedPolygon(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_GetInfo_Area_All(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void File_GetInfo_Position_SelectedVertex(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	void File_GetInfo_DebugInfo_Selection(FsGuiPopUpMenuItem *);
	void File_GetInfo_DebugInfo_Selection_AskedCopyToClipboard(FsGuiDialog *dlg,int returnCode);
	void File_GetInfo_DebugInfo_Selection_ShowOnDialog(void);
	YsWString File_GetInfo_DebugInfo_Selection_MakeMessage(YSBOOL first20) const;

	static void File_Recent(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);



	static void Edit_ClearUIMode(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	void Edit_ClearUIMode(void);
	void Edit_ClearUIModeExceptDNMEditMode(void);

	static void Edit_Undo(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void Edit_Redo(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);

	static void Edit_Copy(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void Edit_Paste(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static void Edit_PasteSpecial(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);

	static void Edit_DeleteSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_DeleteSelection(void);

	static void Edit_ForceDeleteSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_InsertVertex(void *appPtr,FsGuiPopUpMenu *menu,FsGuiPopUpMenuItem *);
	static YSRESULT Edit_InsertVertex_SpaceKeyCallBack(GeblGuiEditorBase &canvas);

	static void Edit_InsertPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_InsertConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_InsertFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_SmoothShadeSelectedVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_FlatShadeSelectedVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Paint_ByPickingPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_Paint_ByPickingPolygon_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	static void Edit_Paint_ByPickingFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Paint_SelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Paint_All(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Paint_FloodFill_SameColor(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Edit_Paint_FloodFill_SameColor_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	static void Edit_Paint_FloodFill_BoundedByConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Edit_Paint_FloodFill_BoundedByConstEdge_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);


	static void Edit_CreatePrimitiveShape(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Edit_CreatePrimitive_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	static void Edit_CreatePrimitive_CursorMoved(void *appPtr,FsGui3DInterface &);
	static void Edit_CreatePrimitive_UIDrawCallBack3D(GeblGuiEditorBase &canvas);


	static void Edit_CreateEquilateral(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_CreateEquilateral_DrawCallBack3D(GeblGuiEditorBase &canvas);
	static void Edit_CreateEquilateral_3DInterfaceCallBack(void *appPtr,class FsGui3DInterface &itfc);
	static YSRESULT Edit_CreateEquilateral_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_CreateEquilateral_SpaceKeyCallBack(void);


	static void Edit_CreateAirfoil(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_CreateAirfoil_DrawCallBack3D(GeblGuiEditorBase &canvas);
	void Edit_CreateAirfoil_DrawCallBack3D(void);
	static void Edit_CreateAirfoil_PivotMoved(void *appPtr,FsGui3DInterface &itfc);
	void Edit_CreateAirfoil_Create(void);

	static void Edit_CreateMirrorImage(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_CreateMirrorImage_Create(void);

	static void Edit_CreateParallelogramFromTwoVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_CreateParallelogramFromTwoVertex_DrawCallBack3D(GeblGuiEditorBase &canvas);
	static void Edit_CreateParallelogramFromTwoVertex_3DInterfaceCallBack(void *appPtr,class FsGui3DInterface &itfc);
	static YSRESULT Edit_CreateParallelogramFromTwoVertex_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_CreateParallelogramFromTwoVertex_SpaceKeyCallBack(void);

	static void Edit_CreateParallelogramFromThreeVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_CreateSolidOfRevolution(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_CreateSolidOfRevolution_DrawCallBack3D(GeblGuiEditorBase &canvas);
	static void Edit_CreateSolidOfRevolution_3DInterfaceCallBack(void *appPtr,class FsGui3DInterface &itfc);
	static YSRESULT Edit_CreateSolidOfRevolution_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_CreateSolidOfRevolution_SpaceKeyCallBack(void);

	void Edit_SewBetweenTwoVertex(FsGuiPopUpMenuItem *);


	static void Edit_MoveSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_MoveSelection_DrawCallBack3D(GeblGuiEditorBase &canvas);
	static void Edit_MoveSelection_CursorMoved(void *appPtr,FsGui3DInterface &itfc);
	static YSRESULT Edit_MoveSelection_SpaceKeyCallBack(GeblGuiEditorBase &canvas);

	static void Edit_RotateSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_RotateSelection_DrawCallBack3D(GeblGuiEditorBase &canvas);
	static void Edit_RotateSelection_CursorMoved(void *appPtr,FsGui3DInterface &itfc);
	static YSRESULT Edit_RotateSelection_SpaceKeyCallBack(GeblGuiEditorBase &canvas);

	static void Edit_SwapEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_MergePolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_SplitPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_VertexCollapseMidPoint(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_VertexCollapseAtoB(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_VertexCollapse(YsShellVertexHandle vtHdToDel,YsShellVertexHandle vtHdToSurvive,const YsVec3 &newPos);

	static void Edit_Hemming_AroundSelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Hemming_AroundSelectedPolygon_UIDrawCallBack3D(GeblGuiEditorBase &canvas);

	static void Edit_Hemming_AroundOutsideOfSelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Hemming_OneSideOfVertexSequence(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	void Edit_Hemming_AroundSelectedPolygon(void);

	static void Edit_Intersection_CutByPlane(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Intersection_ChopOff(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Intersection_ChopOff_DrawCallBack3D(GeblGuiEditorBase &canvas);
	static void Edit_Intersection_ChopOff_3DInterfaceCallBack(void *appPtr,class FsGui3DInterface &itfc);
	static YSRESULT Edit_Intersection_ChopOff_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_Intersection_ChopOff_SpaceKeyCallBack(void);

	static void Edit_Sweep_ParallelSweep(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Sweep_ParallelSweep_DrawCallBack3D(GeblGuiEditorBase &canvas);
	void Edit_Sweep_ParallelSweep_DrawCallBack3D(void);
	static YSRESULT Edit_Sweep_ParallelSweep_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_Sweep_ParallelSweep_SpaceKeyCallBack(void);

	static void Edit_Sweep_ParallelSweepWithPath(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Sweep_ParallelSweepWithPath_DrawCallBack3D(GeblGuiEditorBase &canvas);
	void Edit_Sweep_ParallelSweepWithPath_DrawCallBack3D(void);
	static YSRESULT Edit_Sweep_ParallelSweepWithPath_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_Sweep_ParallelSweepWithPath_SpaceKeyCallBack(void);

	static void Edit_Sweep_AlongSweepPath(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Sweep_BetweenConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Sweep_BetweenConstEdge_DrawCallBack3D(GeblGuiEditorBase &canvas);
	static YSRESULT Edit_Sweep_BetweenConstEdge_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_Sweep_BetweenConstEdge_Sweep(void);

	static void Edit_Sweep_ExtendOrShrink(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Sweep_ExtendOrShrink_DrawCallBack3D(GeblGuiEditorBase &canvas);
	static void Edit_Sweep_ExtendOrShrink_3DInterfaceCallBack(void *appPtr,class FsGui3DInterface &itfc);
	static YSRESULT Edit_Sweep_ExtendOrShrink_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_Sweep_ExtendOrShrink_SpaceKeyCallBack(void);

	static void Edit_Offset_Thickening(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Offset_Thickening_DrawCallBack3D(GeblGuiEditorBase &canvas);
	static YSRESULT Edit_Offset_Thickening_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_Offset_Thickening_SpaceKeyCallBack(void);

	static void Edit_Round_PolygonOrConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_Round_PolygonOrConstEdge(void);

	static void Edit_Round_PolygonOrConstEdgeSelectedCorner(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_Round_PolygonOrConstEdgeSelectedCorner(void);

	static void Edit_Round_DrawCallBack2d(GeblGuiEditorBase &canvas);
	void Edit_Round_DrawCallBack2d(void);

	YSRESULT Edit_Round2d(void);

	static void Edit_Round_AroundVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Edit_Round_AroundVertex(const double radius,const int nDiv);

	static void Edit_Round_AroundSelectedPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Round_OpenVertexSequence(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_Round_DrawCallBack3d(GeblGuiEditorBase &canvas);
	void Edit_Round_DrawCallBack3d(void);
	static YSRESULT Edit_Round_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_Round_SpaceKeyCallBack(void);

	YSRESULT Edit_Round3d(void);


	static void Edit_Projection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Edit_Projection_DrawCallBack3d(GeblGuiEditorBase &canvas);
	static YSRESULT Edit_Projection_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT Edit_Projection_Project(void);

	static void Edit_TriangulateSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_InvertConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_SplitConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Edit_MergeConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);



	static void BoolUnion(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolSubtraction(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolIntersection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolMerge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolSeparate(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolOpenSurfaceMinusSolid(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolOpenSurfacePlusSolid(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void BoolOpenSurfaceAndSolid(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Bool_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	void Bool_DiagnoseDialog(void);
	static void Bool_DiagnoseDialogCallBack(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);
	YSRESULT Bool_Apply(YSBOOLEANOPERATION boolOpType,YsShellExtEdit &shlA,YsShellExtEdit &shlB);



	static void Select_UnselectAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PickVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_PickVertex_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	static YSRESULT Select_PickVertex_LButtonBoxCallBack(GeblGuiEditorBase &canvas,int xMin,int yMin,int xMax,int yMax);
	static YSRESULT Select_PickVertex_BackSpaceKeyCallBack(GeblGuiEditorBase &canvas);

	static void Select_PickPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_PickPolygon_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	static YSRESULT Select_PickPolygon_LButtonBoxCallBack(GeblGuiEditorBase &canvas,int xMin,int yMin,int xMax,int yMax);
	static YSRESULT Select_PickPolygon_BackSpaceKeyCallBack(GeblGuiEditorBase &canvas);

	static void Select_PickConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_PickConstEdge_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	static YSRESULT Select_PickConstEdge_LButtonBoxCallBack(GeblGuiEditorBase &canvas,int xMin,int yMin,int xMax,int yMax);
	static YSRESULT Select_PickConstEdge_BackSpaceKeyCallBack(GeblGuiEditorBase &canvas);

	static void Select_PickFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_PickFaceGroup_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	static YSRESULT Select_PickFaceGroup_BackSpaceKeyCallBack(GeblGuiEditorBase &canvas);


	static void Select_UnselectVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_AllVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_InvertVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_VertexOfPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_VertexOfConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_VertexAlongBoundaryOrConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_Vertex_SingleUseEdgeLoop(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	void Select_EdgeConnectedVertexBetweenSelection(FsGuiPopUpMenuItem *);

	static void Select_UnselectPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_AllPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_InvertPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PolygonOfFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PolygonConnected(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PolygonFromEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PolygonUsingAtLeastOneOfSelectedVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_AllFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_InvertFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	void Select_FaceGroup_PlanarFaceGroup(FsGuiPopUpMenuItem *);


	static void Select_CurrentShellByPicking(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT Select_CurrentShellByPicking_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	static void Select_NextShell(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Select_PrevShell(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	void Select_ByProperty(FsGuiPopUpMenuItem *);
	void Select_VertexByHashKey(FsGuiPopUpMenuItem *);
	void Select_VertexByIndex(FsGuiPopUpMenuItem *);
	void Select_PolygonByHashKey(FsGuiPopUpMenuItem *);
	void Select_PolygonByIndex(FsGuiPopUpMenuItem *);
	void Select_FaceGroupByIdNumber(FsGuiPopUpMenuItem *);
	void Select_FaceGroupByLabel(FsGuiPopUpMenuItem *);
	void Select_ConstEdgeByIdNumber(FsGuiPopUpMenuItem *);
	void Select_ConstEdgeByLabel(FsGuiPopUpMenuItem *);
	class PolyCreSelectByPropertyDialog *PrepareSelectByPropDialog(void);
	void SelectTangledPolygon(FsGuiPopUpMenuItem *);

	void View_RefreshBuffer(FsGuiPopUpMenuItem *);

	static void View_DrawAxis(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawPolygonFill(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawWireframe(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawShrunkPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_DrawBackFaceInDifferentColor(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_FocusOn_SelectionCenterOfGravity(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_FocusOn_SelectionBoundingBoxCenter(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_FocusOn_Vertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_FocusOn_Vertex_RKey(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT View_FocusOn_Vertex_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT View_FocusOn_Vertex_Click(int mx,int my);

	static void View_FocusOn_Polygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_FocusOn_Polygon_ShiftRKey(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT View_FocusOn_Polygon_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	static void View_FocusOn_PolygonStraight(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_FocusOn_Polygon_ShiftCtrlRKey(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static YSRESULT View_FocusOn_PolygonStraight_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	YSRESULT View_FocusOn_Polygon_Click(int mx,int my,YSBOOL alsoChangeOrientation);

	void View_ApplyViewPortToAllShell(void);

	static void View_UseSameViewPortForAllShell(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_EnableViewPort(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_ResetViewPort(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_MoveViewPort(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_MoveViewPort_BoxMoved(void *appPtr,FsGui3DInterface &itfc);

	static void View_MoveViewPortToRotationCenter(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_ExpandViewPort(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_ShrinkViewPort(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_ViewPortFromSelection(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_CrossSection_Enable(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void View_CrossSection_Reset(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_CrossSection_Move(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_CrossSection_Invert(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);


	static void View_OpenRefBmpDialog(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_OpenRefBmpDialog_UIDrawCallBack3D(GeblGuiEditorBase &canvas);

	static void View_LoadRefBmpConfig(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_LoadRefBmpConfig_FileSelected(FsGuiControlBase *canvas,FsGuiDialog *,int);

	static void View_SaveRefBmpConfig(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_SaveRefBmpConfig_FileSelected(FsGuiControlBase *canvas,FsGuiDialog *,int);
	static void View_SaveRefBmpConfig_PassedOverwriteCheck(FsGuiControlBase *canvas,FsGuiDialog *,int);
	void View_SaveRefBmpConfig_PassedOverwriteCheck(const wchar_t fn[]);

	static void View_ShowAllRefBmp(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_HideAllRefBmp(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	void View_Highlight_ShowHighlight(FsGuiPopUpMenuItem *);
	static void View_HighlightNonManifoldEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void View_HighlightSelfIntersectingPolygon(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void View_Highlight_HighAspectRatioTriangleAndQuadrilateral(FsGuiPopUpMenuItem *);
	void View_Highlight_HighAspectRatioTriangleAndQuadrilateral_RefreshCallBack(FsGuiDialog *dlgPtr);


	YsArray <YsShellDnmContainer <YsShellExtEditGui>::Node *> Global_GetTarget(void);

	void Global_Target_Current(FsGuiPopUpMenuItem *);
	void Global_Target_All(FsGuiPopUpMenuItem *);

	static void Global_DeleteUnusedVertexAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Global_InvertPolygonAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Global_SetZeroNormalAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Global_FixOrientationAndNormalClosedVolume(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	void Global_FixOrientationBasedOnAssignedNormal(FsGuiPopUpMenuItem *);

	static void Global_TriangulateAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void Global_DeleteRedundantVertex(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Global_DeleteRedundantVertex_ToleranceEntered(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);

	static void Global_DeleteRedundantVertexNonManifoldEdgeOnly(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Global_DeleteRedundantVertexNonManifoldEdgeOnly_ToleranceEntered(FsGuiControlBase *appPtr,FsGuiDialog *,int returnValue);

	void Global_DeleteRedundantVertex(const double tol,const YsShellVertexStore *limitedVtx);



	static void Sketch_CreateConstEdgeBySketch(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Sketch_DrawCallBack2D(GeblGuiEditorBase &canvas);
	static YSRESULT Sketch_CreateConstEdge_LButtonCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	static YSRESULT Sketch_CreateConstEdge_MouseMoveCallBack(GeblGuiEditorBase &canvas,YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);



	static void UtilMenu_ClearConstEdgeAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void UtilMenu_AddConstEdgeByDihedralAngle(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void UtilMenu_AddConstEdgeByDihedralAngle_DrawCallBack3D(GeblGuiEditorBase &canvas);
	static YSRESULT UtilMenu_AddConstEdgeByDihedralAngle_SpaceKeyCallBack(GeblGuiEditorBase &canvas);
	YSRESULT UtilMenu_AddConstEdgeByDihedralAngle_SpaceKeyCallBack(void);

	static void UtilMenu_AddConstEdgeAlongNonManifoldEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void UtilMenu_AddConstEdgeOnFaceGroupBoundary(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void UtilMenu_OpenCloseConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void UtilMenu_ReconsiderConstEdge(FsGuiPopUpMenuItem *);

	static void UtilMenu_ClearFaceGroupAll(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void UtilMenu_ConstrainAllEdges(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void UtilMenu_MakeFaceGroupFromConstEdge(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void UtilMenu_PaintBasedOnFaceGroup(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void UtilMenu_FaceGroup_SetName(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);



	static void RepairMenu_FreeStitching(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);

	static void RepairMenu_LoopStitching(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);



	static void Dnm_LoadStaticMode(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Dnm_LoadState0(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Dnm_LoadState1(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Dnm_LoadState2(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Dnm_LoadState3(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Dnm_Preview(FsGuiPopUpMenuItem *);



	static void Option_Config(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	void Option_ReapplyConfig(class PolyCreConfig &cfg);
	void Option_ReapplyConfig(void);



	static void Help_About(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	static void Help_Help(void *appPtr,FsGuiPopUpMenu *,FsGuiPopUpMenuItem *);
	// Menu Call Back Functions <<
};

extern void FsGui3DSetAppMustTerminate(YSBOOL sw);

/* } */
#endif
