/* ////////////////////////////////////////////////////////////

Polygon Crest
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

File Name: ysgebl_gui_foundation_draw.cpp
//////////////////////////////////////////////////////////// */

#include <ysgl.h>


#include "ysgebl_gui_foundation.h"



void GeblGuiFoundation::DrawAxis(const double length,const int viewport[4],const YsMatrix4x4 &projMat,const YsMatrix4x4 &viewMat) const
{
	const GLfloat axisLength=(GLfloat)length;
	const GLfloat axisLine[6*3]=
	{
		0,0,0,  axisLength,0,0,
		0,0,0,  0,axisLength,0,
		0,0,0,  0,0,axisLength
	};
	const GLfloat axisColor[4]=
	{
		0,0,0,1.0f
	};

	struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
	YsGLSLUse3DRenderer(renderer);

	YsGLSLSet3DRendererUniformColorfv(renderer,axisColor);
	YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,6,axisLine);

	int nTri=0;
	GLfloat triCoord[3*3*3];
	YsVec3 tri[3];
	if(YSOK==FsGui3DGetArrowHeadTriangle(tri,YsOrigin(),axisLength*YsXVec(),viewport,projMat,viewMat,0.1,10,8))
	{
		triCoord[nTri*9  ]=(GLfloat)tri[0].x();
		triCoord[nTri*9+1]=(GLfloat)tri[0].y();
		triCoord[nTri*9+2]=(GLfloat)tri[0].z();
		triCoord[nTri*9+3]=(GLfloat)tri[1].x();
		triCoord[nTri*9+4]=(GLfloat)tri[1].y();
		triCoord[nTri*9+5]=(GLfloat)tri[1].z();
		triCoord[nTri*9+6]=(GLfloat)tri[2].x();
		triCoord[nTri*9+7]=(GLfloat)tri[2].y();
		triCoord[nTri*9+8]=(GLfloat)tri[2].z();
		++nTri;
	}
	if(YSOK==FsGui3DGetArrowHeadTriangle(tri,YsOrigin(),axisLength*YsYVec(),viewport,projMat,viewMat,0.1,10,8))
	{
		triCoord[nTri*9  ]=(GLfloat)tri[0].x();
		triCoord[nTri*9+1]=(GLfloat)tri[0].y();
		triCoord[nTri*9+2]=(GLfloat)tri[0].z();
		triCoord[nTri*9+3]=(GLfloat)tri[1].x();
		triCoord[nTri*9+4]=(GLfloat)tri[1].y();
		triCoord[nTri*9+5]=(GLfloat)tri[1].z();
		triCoord[nTri*9+6]=(GLfloat)tri[2].x();
		triCoord[nTri*9+7]=(GLfloat)tri[2].y();
		triCoord[nTri*9+8]=(GLfloat)tri[2].z();
		++nTri;
	}
	if(YSOK==FsGui3DGetArrowHeadTriangle(tri,YsOrigin(),axisLength*YsZVec(),viewport,projMat,viewMat,0.1,10,8))
	{
		triCoord[nTri*9  ]=(GLfloat)tri[0].x();
		triCoord[nTri*9+1]=(GLfloat)tri[0].y();
		triCoord[nTri*9+2]=(GLfloat)tri[0].z();
		triCoord[nTri*9+3]=(GLfloat)tri[1].x();
		triCoord[nTri*9+4]=(GLfloat)tri[1].y();
		triCoord[nTri*9+5]=(GLfloat)tri[1].z();
		triCoord[nTri*9+6]=(GLfloat)tri[2].x();
		triCoord[nTri*9+7]=(GLfloat)tri[2].y();
		triCoord[nTri*9+8]=(GLfloat)tri[2].z();
		++nTri;
	}

	if(0<nTri)
	{
		YsGLSLDrawPrimitiveVtxfv(renderer,GL_TRIANGLES,nTri*3,triCoord);
	}


	YsGLSLEndUse3DRenderer(renderer);
}



void GeblGuiFoundation::DrawVertex(const YsShellExtEdit &shl,const int viewport[4]) const
{
	struct YsGLSL3DRenderer *renderer=YsGLSLSharedMonoColorMarker3DRenderer();
	YsShellExtDrawingBuffer &buf=shl.GetDrawingBuffer();
	YsGLSLUse3DRenderer(renderer);

	const GLfloat vtxCol[4]={1.0,0.0,1.0,1.0};
	YsGLSLSet3DRendererUniformColorfv(renderer,vtxCol);
	if(0<buf.normalVertexPosBuffer.GetNumVertex())
	{
		YsGLSLSet3DRendererUniformMarkerDimension(renderer,normalVertexMarkerRadius);
		YsGLSLSet3DRendererUniformMarkerType(renderer,YSGLSL_MARKER_TYPE_RECT);
		YsGLSLSet3DRendererViewportDimensionf(renderer,(GLfloat)viewport[2],(GLfloat)viewport[3]);
		YsGLSLDrawPrimitiveVtxPixelOffsetfv(renderer,GL_TRIANGLES,buf.normalVertexPosBuffer.GetNumVertex(),buf.normalVertexPosBuffer,buf.normalVertexPixOffsetBuffer);
	}
	if(0<buf.roundVertexPosBuffer.GetNumVertex())
	{
		YsGLSLSet3DRendererUniformMarkerDimension(renderer,roundVertexMarkerRadius);
		YsGLSLSet3DRendererUniformMarkerType(renderer,YSGLSL_MARKER_TYPE_CIRCLE);
		YsGLSLSet3DRendererViewportDimensionf(renderer,(GLfloat)viewport[2],(GLfloat)viewport[3]);
		YsGLSLDrawPrimitiveVtxPixelOffsetfv(renderer,GL_TRIANGLES,buf.roundVertexPosBuffer.GetNumVertex(),buf.roundVertexPosBuffer,buf.roundVertexPixOffsetBuffer);
	}

	YsGLSLEndUse3DRenderer(renderer);
}



void GeblGuiFoundation::DrawSelectedVertex(const YsShellExtEdit &shl,const int viewport[4]) const
{
	struct YsGLSL3DRenderer *renderer=YsGLSLSharedMonoColorMarker3DRenderer();
	YsShellExtDrawingBuffer &buf=shl.GetDrawingBuffer();
	YsGLSLUse3DRenderer(renderer);

	const GLfloat selVtxCol[4]={0.0f,0.0f,1.0f,1.0f};
	YsGLSLSet3DRendererUniformColorfv(renderer,selVtxCol);
	YsGLSLSet3DRendererUniformMarkerDimension(renderer,selectedVertexMarkerRadius);
	YsGLSLSet3DRendererUniformMarkerType(renderer,YSGLSL_MARKER_TYPE_EMPTY_RECT);
	YsGLSLSet3DRendererViewportDimensionf(renderer,(GLfloat)viewport[2],(GLfloat)viewport[3]);
	YsGLSLDrawPrimitiveVtxPixelOffsetfv(renderer,GL_TRIANGLES,buf.selectedVertexPosBuffer.GetNumVertex(),buf.selectedVertexPosBuffer,buf.selectedVertexPixOffsetBuffer);

	YsGLSLEndUse3DRenderer(renderer);
}



void GeblGuiFoundation::DrawSelectedVertexLine(const YsShellExtEdit &shl) const
{
	struct YsGLSL3DRenderer *renderer=YsGLSLSharedVariColor3DRenderer();
	YsShellExtDrawingBuffer &buf=shl.GetDrawingBuffer();
	YsGLSLUse3DRenderer(renderer);
	YsGLSLDrawPrimitiveVtxColfv(renderer,GL_LINES,buf.selectedVertexLineBuffer.GetNumVertex(),buf.selectedVertexLineBuffer,buf.selectedVertexLineColBuffer);
	YsGLSLEndUse3DRenderer(renderer);
}



void GeblGuiFoundation::DrawPolygonEdge(const YsShellExtEdit &shl) const
{
	struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
	YsShellExtDrawingBuffer &buf=shl.GetDrawingBuffer();
	YsGLSLUse3DRenderer(renderer);

	const GLfloat edgeCol[4]={0,0,0,1.0f};
	YsGLSLSet3DRendererUniformColorfv(renderer,edgeCol);
	YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,buf.normalEdgePosBuffer.GetNumVertex(),buf.normalEdgePosBuffer);

	YsGLSLEndUse3DRenderer(renderer);
}



void GeblGuiFoundation::DrawSelectedPolygon(const YsShellExtEdit &shl) const
{
	YsGLSL3DRenderer *renderer=YsGLSLSharedVariColor3DRenderer();
	YsShellExtDrawingBuffer &buf=shl.GetDrawingBuffer();
	YsGLSLUse3DRenderer(renderer);
	glLineWidth(3.0f);

	YsGLSLDrawPrimitiveVtxColfv(renderer,GL_LINES,buf.selectedPolygonPosBuffer.GetNumVertex(),buf.selectedPolygonPosBuffer,buf.selectedPolygonColBuffer);

	glLineWidth(1.0f);
	YsGLSLEndUse3DRenderer(renderer);
}



void GeblGuiFoundation::DrawShrunkPolygon(const YsShellExtEdit &shl) const
{
	struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
	YsShellExtDrawingBuffer &buf=shl.GetDrawingBuffer();
	YsGLSLUse3DRenderer(renderer);

	const GLfloat edgeCol[4]={0,0,0,1.0f};
	YsGLSLSet3DRendererUniformColorfv(renderer,edgeCol);
	YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,buf.shrunkEdgePosBuffer.GetNumVertex(),buf.shrunkEdgePosBuffer);

	YsGLSLEndUse3DRenderer(renderer);
}



void GeblGuiFoundation::DrawConstEdge(const YsShellExtEdit &shl) const
{
	struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
	YsShellExtDrawingBuffer &buf=shl.GetDrawingBuffer();
	YsGLSLUse3DRenderer(renderer);
	glLineWidth(3);

	const GLfloat constEdgeCol[4]={0.0,0.0,0.0,1.0};
	YsGLSLSet3DRendererUniformColorfv(renderer,constEdgeCol);
	YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,buf.constEdgeVtxBuffer.GetNumVertex(),buf.constEdgeVtxBuffer);

	glLineWidth(1);
	YsGLSLEndUse3DRenderer(renderer);
}



void GeblGuiFoundation::DrawSelectedConstEdge(const YsShellExtEdit &shl) const
{
	struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
	YsShellExtDrawingBuffer &buf=shl.GetDrawingBuffer();
	YsGLSLUse3DRenderer(renderer);
	glLineWidth(4);

	const GLfloat selectedConstEdgeCol[4]={1.0,0.0,1.0,1.0};
	YsGLSLSet3DRendererUniformColorfv(renderer,selectedConstEdgeCol);
	YsGLSLDrawPrimitiveVtxfv(renderer,GL_LINES,buf.selectedConstEdgeVtxBuffer.GetNumVertex(),buf.selectedConstEdgeVtxBuffer);
	YsGLSLDrawPrimitiveVtxfv(renderer,GL_TRIANGLES,buf.selectedConstEdgeArrowHeadVtxBuffer.GetNumVertex(),buf.selectedConstEdgeArrowHeadVtxBuffer);

	glLineWidth(1);
	YsGLSLEndUse3DRenderer(renderer);
}



void GeblGuiFoundation::DrawSelectedFaceGroup(const YsShellExtEdit &shl) const
{
	struct YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
	YsShellExtDrawingBuffer &buf=shl.GetDrawingBuffer();
	YsGLSLUse3DRenderer(renderer);

	const GLfloat selectedFaceGroupCol[4]={0.0,1.0,1.0,1.0};
	YsGLSLSet3DRendererUniformColorfv(renderer,selectedFaceGroupCol);
	YsGLSLDrawPrimitiveVtxfv(renderer,GL_TRIANGLES,buf.selectedFaceGroupVtxBuffer.GetNumVertex(),buf.selectedFaceGroupVtxBuffer);

	YsGLSLEndUse3DRenderer(renderer);
}
