/* ////////////////////////////////////////////////////////////

Polygon Crest
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

File Name: ysshellext_duplicateutil.h
//////////////////////////////////////////////////////////// */

#ifndef YSSHELLEXT_DUPLICATEUTIL_IS_INCLUDED
#define YSSHELLEXT_DUPLICATEUTIL_IS_INCLUDED
/* { */

#include "ysshellext_mappingutil.h"

class YsShellExt_DuplicateUtil
{
protected:
	const class YsShellExt *srcShl;

private:
	YsShellExt_Mapping <YsShellVertexHandle,YSSIZE_T> vtHdMap;
	YsShellExt_Mapping <YsShellPolygonHandle,YSSIZE_T> plHdMap;
	YsShellExt_Mapping <YsShellExt::ConstEdgeHandle,YSSIZE_T> ceHdMap;
	YsShellExt_Mapping <YsShellExt::FaceGroupHandle,YSSIZE_T> fgHdMap;
	YsShellExt_Mapping <YsShellExt::VolumeHandle,YSSIZE_T> vlHdMap;

public:
	class Vertex
	{
	public:
		YsShellVertexHandle srcVtHd,newVtHd;
		YsVec3 pos;
		YsShellExt::VertexAttrib attrib;
		YSBOOL derived;
	};
	class Polygon
	{
	public:
		YsShellPolygonHandle srcPlHd,newPlHd;

		YsArray <YSSIZE_T> vertexArray;
		YsShellExt::PolygonAttrib attrib;
		YsColor col;
		YsVec3 nom;
		YSBOOL derived;
	};
	class ConstEdge
	{
	public:
		YsShellExt::ConstEdgeHandle srcCeHd,newCeHd;
		YSBOOL isLoop;
		YsArray <YSSIZE_T> vertexArray;
		YsShellExt::ConstEdge attrib;
		YSBOOL derived;
	};
	class FaceGroup
	{
	public:
		YsShellExt::FaceGroupHandle srcFgHd,newFgHd;
		YsArray <YSSIZE_T> polygonArray;
		YsShellExt::FaceGroup attrib;
		YSBOOL derived;
	};
	class Volume
	{
	public:
		YsShellExt::VolumeHandle srcVlHd,newVlHd;
		YsArray <YSSIZE_T> faceGroupArray;
		YsShellExt::Volume attrib;
		YSBOOL derived;
	};

	YsArray <Vertex> vertexArray;
	YsArray <Polygon> polygonArray;
	YsArray <ConstEdge> constEdgeArray;
	YsArray <FaceGroup> faceGroupArray;
	YsArray <Volume> volumeArray;

	YsShellExt_DuplicateUtil();

	YsShellExt_DuplicateUtil(const YsShellExt_DuplicateUtil &incoming);
	YsShellExt_DuplicateUtil &operator=(const YsShellExt_DuplicateUtil &incoming);
	YsShellExt_DuplicateUtil &CopyFrom(const YsShellExt_DuplicateUtil &incoming);

	YsShellExt_DuplicateUtil(YsShellExt_DuplicateUtil &&incoming);
	YsShellExt_DuplicateUtil &operator=(YsShellExt_DuplicateUtil &&incoming);
	YsShellExt_DuplicateUtil &MoveFrom(YsShellExt_DuplicateUtil &incoming);

	void CleanUp(void);

	// Step 1
	void SetShell(const YsShellExt &shl);

	// Step 2
	void AddVertex(YSSIZE_T n,const YsShellVertexHandle hdArray[]);
	template <class T>
	void AddVertex(const T &hdStore);
private:
	YSSIZE_T AddVertex(YsShellVertexHandle vtHd,YSBOOL derived);

public:
	// Step 3
	void AddPolygon(YSSIZE_T n,const YsShellPolygonHandle hdArray[]);
	template <class T>
	void AddPolygon(const T &hdStore);
private:
	YSSIZE_T AddPolygon(YsShellPolygonHandle plHd,YSBOOL derived);

public:
	// Step 4
	void AddConstEdge(YSSIZE_T n,const YsShellExt::ConstEdgeHandle hdArray[]);
	template <class T>
	void AddConstEdge(const T &hdStore);
private:
	YSSIZE_T AddConstEdge(YsShellExt::ConstEdgeHandle ceHd,YSBOOL derived);

public:
	// Step 5
	void AddFaceGroup(YSSIZE_T n,const YsShellExt::FaceGroupHandle hdArray[]);
	template <class T>
	void AddFaceGroup(const T &hdStore);
private:
	YSSIZE_T AddFaceGroup(YsShellExt::FaceGroupHandle fgHd,YSBOOL derived);

public:
	// Step 6
	void AddVolume(YSSIZE_T n,const YsShellExt::VolumeHandle hdArray[]);
	template <class T>
	void AddVolume(const T &hdStore);
private:
	YSSIZE_T AddVolume(YsShellExt::VolumeHandle vlHd,YSBOOL derived);


public:
	YsArray <YsShellVertexHandle> GetNonDerivedNewVertex(void) const;
	YsArray <YsShellPolygonHandle> GetNonDerivedNewPolygon(void) const;
	YsArray <YsShellExt::ConstEdgeHandle> GetNonDerivedNewConstEdge(void) const;
	YsArray <YsShellExt::FaceGroupHandle> GetNonDerivedNewFaceGroup(void) const;
	YsArray <YsShellExt::VolumeHandle> GetNonDerivedNewVolume(void) const;
};


template <class T>
void YsShellExt_DuplicateUtil::AddVertex(const T &hdStore)
{
	for(auto vtHd : hdStore)
	{
		AddVertex(vtHd,YSFALSE);
	}
}

template <class T>
void YsShellExt_DuplicateUtil::AddPolygon(const T &hdStore)
{
	for(auto plHd : hdStore)
	{
		AddPolygon(plHd,YSFALSE);
	}
}

template <class T>
void YsShellExt_DuplicateUtil::AddConstEdge(const T &hdStore)
{
	for(auto ceHd : hdStore)
	{
		AddConstEdge(ceHd,YSFALSE);
	}
}

template <class T>
void YsShellExt_DuplicateUtil::AddFaceGroup(const T &hdStore)
{
	for(auto iter=hdStore.begin(); iter!=hdStore.end(); ++iter)
	{
		YsShellExt::FaceGroupHandle fgHd=*iter;
		AddFaceGroup(fgHd,YSFALSE);
	}
}

template <class T>
void YsShellExt_DuplicateUtil::AddVolume(const T &hdStore)
{
	for(auto vlHd : hdStore)
	{
		AddVolume(vlHd,YSFALSE);
	}
}



/* } */
#endif
