/* ////////////////////////////////////////////////////////////

Polygon Crest
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

File Name: ysshellext_geomcalc.cpp
//////////////////////////////////////////////////////////// */

#include <ysshellext.h>

#include "ysshellext_geomcalc.h"
#include "ysshellext_projectionutil.h"



const double YsShellExt_CalculateDihedralAngle(const YsShellExt &shl,const YsShellPolygonHandle plHd[2])
{
	return YsShellExt_CalculateDihedralAngle(shl,plHd[0],plHd[1]);
}

const double YsShellExt_CalculateDihedralAngle(const YsShellExt &shl,YsShellPolygonHandle plHd0,YsShellPolygonHandle plHd1)
{
	YsArray <YsVec3,4> plVtPos[2];
	shl.GetPolygon(plVtPos[0],plHd0);
	shl.GetPolygon(plVtPos[1],plHd1);

	YsVec3 nom[2];
	if(YSOK==YsGetAverageNormalVector(nom[0],plVtPos[0]) && YSOK==YsGetAverageNormalVector(nom[1],plVtPos[1]))
	{
		return acos(YsBound(nom[0]*nom[1],-1.0,1.0));
	}
	return 0.0;
}



const double YsShellExt_CalculateDihedralAngle(const YsShellExt &shl,const YsShellVertexHandle edVtHd[2])
{
	return YsShellExt_CalculateDihedralAngle(shl,edVtHd[0],edVtHd[1]);
}

const double YsShellExt_CalculateDihedralAngle(const class YsShellExt &shl,const YsShellVertexHandle edVtHd0,const YsShellVertexHandle edVtHd1)
{
	YsArray <YsShellPolygonHandle,2> edPlHd;
	if(YSOK==shl.FindPolygonFromEdgePiece(edPlHd,edVtHd0,edVtHd1) && 2==edPlHd.GetN())
	{
		return YsShellExt_CalculateDihedralAngle(shl,edPlHd);
	}
	return 0.0;
}

const double YsShellExt_CalculateDihedralAngle(const class YsShellExt &shl,const YsShellPolygonHandle plHd,int nThEdgeZeroBased)
{
	YsArray <YsShellVertexHandle,4> plVtHd;
	shl.GetPolygon(plVtHd,plHd);
	return YsShellExt_CalculateDihedralAngle(shl,plVtHd[nThEdgeZeroBased],plVtHd.GetCyclic(nThEdgeZeroBased+1));
}



const double YsShellExt_CalculateTotalDihedralAngleAroundEdge(const class YsShellExt &shl,const YsShellVertexHandle edVtHd[2])
{
	return YsShellExt_CalculateTotalDihedralAngleAroundEdge(shl,edVtHd[0],edVtHd[1]);
}

const double YsShellExt_CalculateTotalDihedralAngleAroundEdge(const class YsShellExt &shl,const YsShellVertexHandle edVtHd0,const YsShellVertexHandle edVtHd1)
{
	YsArray <YsShellPolygonHandle,2> basePlHdPair;
	if(YSOK==shl.FindPolygonFromEdgePiece(basePlHdPair,edVtHd0,edVtHd1) && 2==basePlHdPair.GetN())
	{
		double total=YsShellExt_CalculateDihedralAngle(shl,basePlHdPair);
		for(auto plHd : basePlHdPair)
		{
			YsArray <YsShellVertexHandle,4> plVtHd;
			shl.GetPolygon(plVtHd,plHd);
			for(auto vtIdx : plVtHd.AllIndex())
			{
				if(YSTRUE!=YsSameEdge(plVtHd[vtIdx],plVtHd.GetCyclic(vtIdx+1),edVtHd0,edVtHd1))
				{
					total+=YsShellExt_CalculateDihedralAngle(shl,plVtHd[vtIdx],plVtHd.GetCyclic(vtIdx+1));
				}
			}
		}
		return total;
	}
	return 0.0;
}

const YsVec3 YsShell_CalculatePolygonAverageNormal(const YsShell &shl,YSSIZE_T nPl,const YsShellPolygonHandle plHdArray[])
{
	YsArray <YsVec3,16> allVtPos;
	for(YSSIZE_T plIdx=0; plIdx<nPl; ++plIdx)
	{
		YsArray <YsShellVertexHandle,4> plVtHd;
		shl.GetPolygon(plVtHd,plHdArray[plIdx]);
		for(auto vtHd : plVtHd)
		{
			allVtPos.Append(shl.GetVertexPosition(vtHd));
		}
	}
	YsVec3 nom;
	if(YSOK==YsFindLeastSquareFittingPlaneNormal(nom,allVtPos))
	{
		return nom;
	}
	return YsOrigin();
}

////////////////////////////////////////////////////////////

YsShellExt_NormalCalculator::YsShellExt_NormalCalculator()
{
	weight=NO_WEIGHT;
	limitFgHd=NULL;
}

YsVec3 YsShellExt_NormalCalculator::CalculateVertexNormal(const YsShellExt &shl,YsShellVertexHandle vtHd) const
{
	YsVec3 nomSum=YsOrigin();

	YSSIZE_T nVtPl;
	const YsShellPolygonHandle *vtPlHd;
	if(YSOK==shl.FindPolygonFromVertex(nVtPl,vtPlHd,vtHd))
	{
		for(auto idx=nVtPl-1; 0<=idx; --idx)
		{
			if(NULL!=limitFgHd && shl.FindFaceGroupFromPolygon(vtPlHd[idx])!=limitFgHd)
			{
				continue;
			}

			double weight=1.0;
			nomSum+=weight*shl.GetNormal(vtPlHd[idx]);
		}
		nomSum.Normalize();
	}

	return nomSum;
}

YsVec3 YsShellExt_NormalCalculator::CalculateEdgeNormal(const YsShellExt &shl,const YsShellVertexHandle edVtHd[2]) const
{
	YsVec3 nomSum=YsOrigin();

	YSSIZE_T nEdPl;
	const YsShellPolygonHandle *edPlHd;
	if(YSOK==shl.FindPolygonFromEdgePiece(nEdPl,edPlHd,edVtHd))
	{
		for(auto idx=nEdPl-1; 0<=idx; --idx)
		{
			if(NULL!=limitFgHd && shl.FindFaceGroupFromPolygon(edPlHd[idx])!=limitFgHd)
			{
				continue;
			}

			double weight=1.0;
			nomSum+=weight*shl.GetNormal(edPlHd[idx]);
		}
		nomSum.Normalize();
	}

	return nomSum;
}

void YsShellExt_CalculateQuadricErrorMetric(YsMatrix4x4 &Kp,const YsShellExt &shl,YsShellVertexHandle vtHd)
{
	auto vtPlHd=shl.FindPolygonFromVertex(vtHd);

	Kp.MakeZero();

	const YsVec3 o=shl.GetVertexPosition(vtHd);

	for(auto plHd : vtPlHd)
	{
		auto plVtHd=shl.GetPolygonVertex(plHd);
		YsVec3 n=YsOrigin();

		if(3==plVtHd.GetN())
		{
			const YsVec3 plVtPos[3]=
			{
				shl.GetVertexPosition(plVtHd[0]),
				shl.GetVertexPosition(plVtHd[1]),
				shl.GetVertexPosition(plVtHd[2])
			};
			const YsVec3 edVec[2]=
			{
				plVtPos[1]-plVtPos[0],
				plVtPos[2]-plVtPos[0]
			};
			n=YsUnitVector(edVec[0]^edVec[1]);
		}
		else
		{
			for(int idx=0; idx<plVtHd.GetN(); ++idx)
			{
				if(vtHd==plVtHd[idx])
				{
					const YsVec3 plVtPos[3]=
					{
						shl.GetVertexPosition(plVtHd.GetCyclic(idx-1)),
						shl.GetVertexPosition(plVtHd[idx]),
						shl.GetVertexPosition(plVtHd.GetCyclic(idx+1))
					};
					const YsVec3 edVec[2]=
					{
						plVtPos[1]-plVtPos[0],
						plVtPos[2]-plVtPos[0]
					};
					n=YsUnitVector(edVec[0]^edVec[1]);
					break;
				}
			}
		}

		// (p-o)*n=0 -> p*n-o*n=0  -> nx*x+ny*y+nz*z-o*n=0
		const double a=n.x(),b=n.y(),c=n.z(),d=-o*n;
		const double ab=a*b;
		const double ac=a*c;
		const double ad=a*d;
		const double bc=b*c;
		const double bd=b*d;
		const double cd=c*d;

		YsMatrix4x4 m;
		m.Set(1,1,a*a);
		m.Set(1,2,ab);
		m.Set(1,3,ac);
		m.Set(1,4,ad);

		m.Set(2,1,ab);
		m.Set(2,2,b*b);
		m.Set(2,3,bc);
		m.Set(2,4,bd);

		m.Set(3,1,ac);
		m.Set(3,2,bc);
		m.Set(3,3,c*c);
		m.Set(3,4,cd);

		m.Set(4,1,ad);
		m.Set(4,2,bd);
		m.Set(4,3,cd);
		m.Set(4,4,d*d);

		Kp+=m;
	}
}


double YsShellExt_CalculateQuadricError(const YsMatrix4x4 &Kp,const YsVec3 &pos)
{
	//
	//               v11    v12    v13    v14
	//               v21    v22    v23    v24
	//               v31    v32    v33    v34
	//               v41    v42    v43    v44
	//  [x y z 1]

	const double posTimesKp[4]=
	{
		pos.x()*Kp.v(1,1)+pos.y()*Kp.v(2,1)+pos.z()*Kp.v(3,1)+Kp.v(4,1),
		pos.x()*Kp.v(1,2)+pos.y()*Kp.v(2,2)+pos.z()*Kp.v(3,2)+Kp.v(4,2),
		pos.x()*Kp.v(1,3)+pos.y()*Kp.v(2,3)+pos.z()*Kp.v(3,3)+Kp.v(4,3),
		pos.x()*Kp.v(1,4)+pos.y()*Kp.v(2,4)+pos.z()*Kp.v(3,4)+Kp.v(4,4),
	};

	return pos[0]*posTimesKp[0]+pos[1]*posTimesKp[1]+pos[2]*posTimesKp[2]+posTimesKp[3];
}

double YsShellExt_CalculateTotalFaceAngle(const YsShellExt &shl,YsShellVertexHandle vtHd)
{
	auto vtPlHd=shl.FindPolygonFromVertex(vtHd);

	double totalAngle=0.0;
	for(auto plHd : vtPlHd)
	{
		totalAngle+=YsShellExt_CalculatePolygonFaceAngle(shl,plHd,vtHd);
	}

	return totalAngle;
}

double YsShellExt_CalculatePolygonFaceAngle(const YsShellExt &shl,YsShellPolygonHandle plHd,YsShellVertexHandle vtHd)
{
	return YsShellExt_CalculatePolygonFaceAngle(shl,shl.GetPolygonVertex(plHd),vtHd);
}

double YsShellExt_CalculatePolygonFaceAngle(const YsShellExt &shl,YSSIZE_T nVtIn,const YsShellVertexHandle plVtHdIn[],YsShellVertexHandle vtHd)
{
	YSSIZE_T vtIdx=-1;
	YsArray <YsVec3,4> plVtPos;
	plVtPos.Resize(nVtIn);
	for(YSSIZE_T idx=0; idx<nVtIn; ++idx)
	{
		plVtPos[idx]=shl.GetVertexPosition(plVtHdIn[idx]);
		if(plVtHdIn[idx]==vtHd)
		{
			vtIdx=idx;
		}
	}

	YsShellExt_PolygonProjection prj;
	prj.Project(plVtPos);
	auto prjVtPos=prj.GetProjectedPolygon();

	if(0<=vtIdx)
	{
		return YsGetPolygonFaceAngle(prjVtPos,vtIdx);
	}
	return 0.0;
}

YSBOOL YsShellExt_IsFaceGroupPlanar(const YsShellExt &shl,YsShellExt::FaceGroupHandle fgHd,const double cosineTolerance)
{
	auto fgPlHd=shl.GetFaceGroup(fgHd);

	if(0<fgPlHd.GetN())
	{
		const YsVec3 refNom=shl.GetNormal(fgPlHd[0]);
		for(auto plHd : fgPlHd)
		{
			const double c=refNom*shl.GetNormal(plHd);
			if(c<cosineTolerance)
			{
				return YSFALSE;
			}
		}
		return YSTRUE;
	}
	return YSFALSE;
}

void YsShellExt_CalculateAspectRatio(double &length,double &height,YSSIZE_T np,const YsVec3 p[])
{
	if(3<=np)
	{
		auto longestIdx=YsFindLongestEdgeIndexOfPolygon(np,p);

		const YsVec3 &edVtPos0=p[longestIdx];
		const YsVec3 &edVtPos1=YsGetCyclic(np,p,longestIdx+1);

		length=(edVtPos1-edVtPos0).GetLength();


		height=0;
		for(YSSIZE_T idx=0; idx<np; ++idx)
		{
			if(idx!=longestIdx && idx!=longestIdx+1) // This will miss idx==np-1 and longestIdx=0, but can save bunch of integer divisions.
			{
				const double h=YsGetPointLineDistance3(edVtPos0,edVtPos1,p[idx]);
				YsMakeGreater(height,h);
			}
		}
	}
	else
	{
		length=0.0;
		height=0.0;
	}
}

void YsShellExt_CalculateAspectRatio(double &length,double &height,const YsShell &shl,YSSIZE_T nv,const YsShellVertexHandle v[])
{
	YsArray <YsVec3,4> plg(nv,NULL);
	for(YSSIZE_T idx=0; idx<nv; ++idx)
	{
		plg[idx]=shl.GetVertexPosition(v[idx]);
	}
	YsShellExt_CalculateAspectRatio(length,height,nv,plg);
}

void YsShellExt_CalculateAspectRatio(double &length,double &height,const YsShell &shl,YsShellPolygonHandle plHd)
{
	int nPlVt;
	const YsShellVertexHandle *plVtHd;
	shl.GetPolygon(nPlVt,plVtHd,plHd);
	YsShellExt_CalculateAspectRatio(length,height,shl,nPlVt,plVtHd);
}


