/* ////////////////////////////////////////////////////////////

Polygon Crest
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

File Name: ysshellext_localop.cpp
//////////////////////////////////////////////////////////// */

#include <stdio.h>
#include "ysshellext_localop.h"



void YsShell_LocalOperation::FaceGroupSplit::CleanUp(void)
{
	fgHd=NULL;
	fgPlHdArray.CleanUp();
}



////////////////////////////////////////////////////////////

YsShell_SwapInfo::YsShell_SwapInfo()
{
	CleanUp();
}

void YsShell_SwapInfo::CleanUp(void)
{
	orgDiagonal[0]=NULL;
	orgDiagonal[1]=NULL;
	newDiagonal[0]=NULL;
	newDiagonal[1]=NULL;
	quadVtHd[0]=NULL;
	quadVtHd[1]=NULL;
	quadVtHd[2]=NULL;
	quadVtHd[3]=NULL;
	triPlHd[0]=NULL;
	triPlHd[1]=NULL;

	newTriVtHd[0][0]=NULL;
	newTriVtHd[0][1]=NULL;
	newTriVtHd[0][2]=NULL;
	newTriVtHd[1][0]=NULL;
	newTriVtHd[1][1]=NULL;
	newTriVtHd[1][2]=NULL;
}

YSRESULT YsShell_SwapInfo::MakeInfo(const YsShell &shl,const YsShell::Edge edge)
{
	return MakeInfo(shl,edge.edVtHd[0],edge.edVtHd[1]);
}

YSRESULT YsShell_SwapInfo::MakeInfo(const YsShell &shl,const YsShellVertexHandle edVtHd[2])
{
	return MakeInfo(shl,edVtHd[0],edVtHd[1]);
}

YSRESULT YsShell_SwapInfo::MakeInfo(const YsShell &shl,const YsShellVertexHandle edVtHd0,const YsShellVertexHandle edVtHd1)
{
	const YsShellSearchTable *search=shl.GetSearchTable();
	if(NULL!=search)
	{
		int nEdPl;
		const YsShellPolygonHandle *edPlHd;
		if(YSOK==search->FindPolygonListByEdge(nEdPl,edPlHd,shl,edVtHd0,edVtHd1) && 
		   2==nEdPl &&
		   3==shl.GetNumVertexOfPolygon(edPlHd[0]) &&
		   3==shl.GetNumVertexOfPolygon(edPlHd[1]) &&
		   edPlHd[0]!=edPlHd[1])
		{
			YsArray <YsShellVertexHandle,4> plVtHd0;
			shl.GetVertexListOfPolygon(plVtHd0,edPlHd[0]);

			for(int i=0; i<3 && YSTRUE!=YsSameEdge(plVtHd0[0],plVtHd0[2],edVtHd0,edVtHd1); ++i)
			{
				YsShellVertexHandle swp=plVtHd0[0];
				plVtHd0[0]=plVtHd0[1];
				plVtHd0[1]=plVtHd0[2];
				plVtHd0[2]=swp;
			}

			YsArray <YsShellVertexHandle,4> plVtHd1;
			shl.GetVertexListOfPolygon(plVtHd1,edPlHd[1]);
			for(int i=0; i<3; ++i)
			{
				if(plVtHd1[i]!=edVtHd0 && plVtHd1[i]!=edVtHd1)
				{
					quadVtHd[0]=plVtHd0[0];
					quadVtHd[1]=plVtHd0[1];
					quadVtHd[2]=plVtHd0[2];
					quadVtHd[3]=plVtHd1[i];

					orgDiagonal[0]=edVtHd0;
					orgDiagonal[1]=edVtHd1;

					newDiagonal[0]=quadVtHd[1];
					newDiagonal[1]=quadVtHd[3];

					triPlHd[0]=edPlHd[0];
					triPlHd[1]=edPlHd[1];

					newTriVtHd[0][0]=quadVtHd[1];
					newTriVtHd[0][1]=quadVtHd[2];
					newTriVtHd[0][2]=quadVtHd[3];

					newTriVtHd[1][0]=quadVtHd[3];
					newTriVtHd[1][1]=quadVtHd[0];
					newTriVtHd[1][2]=quadVtHd[1];

					return YSOK;
				}
			}
		}
	}
	return YSERR;
}

YSRESULT YsShell_SwapInfo::MakeInfo(const YsShell &shl,const YsShellPolygonHandle edPlHd[2])
{
	return MakeInfo(shl,edPlHd[0],edPlHd[1]);
}

YSRESULT YsShell_SwapInfo::MakeInfo(const YsShell &shl,const YsShellPolygonHandle edPlHd0,const YsShellPolygonHandle edPlHd1)
{
	if(edPlHd0!=edPlHd1 && 3==shl.GetNumVertexOfPolygon(edPlHd0) && 3==shl.GetNumVertexOfPolygon(edPlHd1))
	{
		YsArray <YsShellVertexHandle,4> plVtHd[2];
		shl.GetVertexListOfPolygon(plVtHd[0],edPlHd0);
		shl.GetVertexListOfPolygon(plVtHd[1],edPlHd1);

		for(int i=0; i<3; ++i)
		{
			for(int j=0; j<3; ++j)
			{
				if(YSTRUE==YsSameEdge(plVtHd[0][i],plVtHd[0].GetCyclic(i+1),plVtHd[1][j],plVtHd[1].GetCyclic(j+1)))
				{
					return MakeInfo(shl,plVtHd[0][i],plVtHd[0].GetCyclic(i+1));
				}
			}
		}
	}
	return YSERR;
}

////////////////////////////////////////////////////////////

YsShell_MergeInfo::YsShell_MergeInfo()
{
	CleanUp();
}

void YsShell_MergeInfo::CleanUp(void)
{
	plHd[0]=NULL;
	plHd[1]=NULL;
	newPlVtHd.CleanUp();
	duplicateWarning=YSFALSE;
}

YSRESULT YsShell_MergeInfo::MakeInfo(const YsShell &shl,const YsShellPolygonHandle plHd[2])
{
	return MakeInfo(shl,plHd[0],plHd[1]);
}

YSRESULT YsShell_MergeInfo::MakeInfo(const YsShell &shl,const YsShellPolygonHandle plHd0,const YsShellPolygonHandle plHd1)
{
	CleanUp();

	YsArray <YsShellVertexHandle> plVtHd[2];
	shl.GetVertexListOfPolygon(plVtHd[0],plHd0);
	shl.GetVertexListOfPolygon(plVtHd[1],plHd1);

	for(int i=0; i<2; ++i)
	{
		printf("[%d]",i);
		for(int j=0; j<plVtHd[i].GetN(); ++j)
		{
			printf(" %d",shl.GetSearchKey(plVtHd[i][j]));
		}
		printf("\n");
	}

	// Find first shared edge
	YSSIZE_T sharedEdIdx[2]={-1,-1};
	for(YSSIZE_T i=0; i<plVtHd[0].GetN(); ++i)
	{
		for(YSSIZE_T j=0; j<plVtHd[1].GetN(); ++j)
		{
			if(plVtHd[0][i]==plVtHd[1][j] && plVtHd[0].GetCyclic(i+1)==plVtHd[1].GetCyclic(j+1))
			{
				sharedEdIdx[0]=i;
				sharedEdIdx[1]=j;
			}
			else if(plVtHd[0][i]==plVtHd[1].GetCyclic(j+1) && plVtHd[0].GetCyclic(i+1)==plVtHd[1][j])
			{
				sharedEdIdx[0]=i;
				plVtHd[1].Invert();
				sharedEdIdx[1]=plVtHd[1].GetN()-1-(j+1)%plVtHd[1].GetN();
			}
		}
	}

	// Validate
	if(0>sharedEdIdx[0] || 0>sharedEdIdx[1])
	{
		return YSERR;
	}

	if(YSTRUE!=YsSameEdge(plVtHd[0][sharedEdIdx[0]],plVtHd[0].GetCyclic(1+sharedEdIdx[0]),
	                      plVtHd[1][sharedEdIdx[1]],plVtHd[1].GetCyclic(1+sharedEdIdx[1])))
	{
		YsPrintf("%s\n",__FUNCTION__);
		YsPrintf("  Internal error.  Shared edge identification failed.\n");
		return YSERR;
	}

	// Rotate
	{
		YsArray <YsShellVertexHandle,16> A,B;
		if(0!=sharedEdIdx[0])
		{
			A.Set(sharedEdIdx[0],plVtHd[0]);
			B.Set(plVtHd[0].GetN()-sharedEdIdx[0],plVtHd[0]+sharedEdIdx[0]);
			plVtHd[0]=B;
			plVtHd[0].Append(A);
		}

		if(0!=sharedEdIdx[1])
		{
			A.Set(sharedEdIdx[1],plVtHd[1]);
			B.Set(plVtHd[1].GetN()-sharedEdIdx[1],plVtHd[1]+sharedEdIdx[1]);
			plVtHd[1]=B;
			plVtHd[1].Append(A);
		}
	}

	// Trim mid shared sequence
	while(3<=plVtHd[0].GetN() &&
	      3<=plVtHd[1].GetN() &&
	      plVtHd[0][0]==plVtHd[1][0] &&
	      plVtHd[0][1]==plVtHd[1][1] &&
	      plVtHd[0][2]==plVtHd[1][2])
	{
		plVtHd[0].Delete(1);
		plVtHd[1].Delete(1);
	}
	while(3<=plVtHd[0].GetN() &&
	      3<=plVtHd[1].GetN() &&
	      plVtHd[0].GetEnd()==plVtHd[1].GetEnd() &&
	      plVtHd[0][0]==plVtHd[1][0] &&
	      plVtHd[0][1]==plVtHd[1][1])
	{
		YsShellVertexHandle swp=plVtHd[0].GetEnd();
		plVtHd[0][0]=swp;
		plVtHd[0].DeleteLast();

		swp=plVtHd[1].GetEnd();
		plVtHd[1][0]=swp;
		plVtHd[1].DeleteLast();
	}

	for(int i=0; i<2; ++i)
	{
		printf("[%d]",i);
		for(int j=0; j<plVtHd[i].GetN(); ++j)
		{
			printf(" %d",shl.GetSearchKey(plVtHd[i][j]));
		}
		printf("\n");
	}

	// Finally assemble merged polygon
	if(2<=plVtHd[0].GetN() && 
	   2<=plVtHd[1].GetN() &&
	   YSTRUE==YsSameEdge(plVtHd[0][0],plVtHd[0][1],plVtHd[1][0],plVtHd[1][1]))
	{
		YsShellVertexHandle swp=plVtHd[0][0];
		plVtHd[0].Append(swp);
		plVtHd[0].Delete(0);

		swp=plVtHd[1][0];
		plVtHd[1].Append(swp);
		plVtHd[1].Delete(0);

		newPlVtHd=plVtHd[0];
		plVtHd[1].Invert();
		newPlVtHd.Append(plVtHd[1].GetN()-2,plVtHd[1]+1);

		for(int i=0; i<newPlVtHd.GetN(); ++i)
		{
			for(int j=i+1; j<newPlVtHd.GetN(); ++j)
			{
				if(newPlVtHd[i]==newPlVtHd[j])
				{
					duplicateWarning=YSTRUE;
				}
			}
		}

		if(3<=newPlVtHd.GetN())
		{
			return YSOK;
		}
	}
	return YSERR;
}

////////////////////////////////////////////////////////////

YsShell_CollapseInfo::YsShell_CollapseInfo()
{
	CleanUp();
}

void YsShell_CollapseInfo::CleanUp(void)
{
	newPlg.Set(0,NULL);
	newConstEdge.Set(0,NULL);
	plHdToDel.Set(0,NULL);
	ceHdToDel.Set(0,NULL);
	vtHdToDel=NULL;
	vtHdToSurvive=NULL;
}

YSRESULT YsShell_CollapseInfo::MakeInfo(const YsShellExt &shl,const YsShellVertexHandle edVtHd[2])
{
	return MakeInfo(shl,edVtHd[0],edVtHd[1]);
}

YSRESULT YsShell_CollapseInfo::MakeInfo(const YsShellExt &shl,const YsShellVertexHandle vtHdToDel,const YsShellVertexHandle vtHdToSurvive)
{
	CleanUp();

	YsArray <YsShellPolygonHandle> vtPlHd;
	if(YSOK==shl.FindPolygonFromVertex(vtPlHd,vtHdToDel))
	{
		for(YSSIZE_T plIdx=0; plIdx<vtPlHd.GetN(); ++plIdx)
		{
			YsArray <YsShellVertexHandle,4> plVtHd;
			shl.GetVertexListOfPolygon(plVtHd,vtPlHd[plIdx]);

			for(auto i=plVtHd.GetN()-1; 0<=i; --i)
			{
				if(plVtHd[i]==vtHdToDel)
				{
					plVtHd[i]=vtHdToSurvive;
				}

				if(plVtHd[i]==plVtHd.GetCyclic(i+1))
				{
					plVtHd.Delete(i);
				}
			}

			if(0<plVtHd.GetN() && plVtHd[0]==plVtHd.Last())
			{
				plVtHd.DeleteLast();
			}

			if(3<=plVtHd.GetN())
			{
				newPlg.Increment();
				newPlg.GetEnd().plHd=vtPlHd[plIdx];
				newPlg.GetEnd().plVtHd=plVtHd;
			}
			else
			{
				plHdToDel.Append(vtPlHd[plIdx]);
			}
		}
	}

	YsArray <YsShellExt::ConstEdgeHandle> vtCeHd;
	if(YSOK==shl.FindConstEdgeFromVertex(vtCeHd,vtHdToDel))
	{
		for(YSSIZE_T ceIdx=0; ceIdx<vtCeHd.GetN(); ++ceIdx)
		{
			YsArray <YsShellVertexHandle,4> ceVtHd;
			YSBOOL isLoop;
			shl.GetConstEdge(ceVtHd,isLoop,vtCeHd[ceIdx]);

			if(YSTRUE!=isLoop)
			{
				for(auto i=ceVtHd.GetN()-1; 0<=i; --i)
				{
					if(ceVtHd[i]==vtHdToDel)
					{
						ceVtHd[i]=vtHdToSurvive;
					}

					if(ceVtHd[i]==ceVtHd.GetCyclic(i+1))
					{
						ceVtHd.Delete(i);
					}
				}

				if(2<=ceVtHd.GetN())
				{
					newConstEdge.Increment();
					newConstEdge.GetEnd().ceHd=vtCeHd[ceIdx];
					newConstEdge.GetEnd().ceVtHd=ceVtHd;
					newConstEdge.GetEnd().isLoop=isLoop;
				}
				else
				{
					ceHdToDel.Append(vtCeHd[ceIdx]);
				}
			}
			else
			{
				for(auto i=ceVtHd.GetN()-1; 0<=i; --i)
				{
					if(ceVtHd[i]==vtHdToDel)
					{
						ceVtHd[i]=vtHdToSurvive;
					}

					if(ceVtHd[i]==ceVtHd.GetCyclic(i+1))
					{
						ceVtHd.Delete(i);
					}
				}

				if(3<=ceVtHd.GetN())
				{
					newConstEdge.Increment();
					newConstEdge.GetEnd().ceHd=vtCeHd[ceIdx];
					newConstEdge.GetEnd().ceVtHd=ceVtHd;
					newConstEdge.GetEnd().isLoop=isLoop;
				}
				else if(2==ceVtHd.GetN())
				{
					newConstEdge.Increment();
					newConstEdge.GetEnd().ceHd=vtCeHd[ceIdx];
					newConstEdge.GetEnd().ceVtHd=ceVtHd;
					newConstEdge.GetEnd().isLoop=YSFALSE;
				}
				else
				{
					ceHdToDel.Append(vtCeHd[ceIdx]);
				}
			}
		}
	}

	this->vtHdToSurvive=vtHdToSurvive;
	this->vtHdToDel=vtHdToDel;

	return YSOK;
}

////////////////////////////////////////////////////////////

YSRESULT YsShellExt_EdgeVertexInsertionInfo::MakeInfo(const YsShellExt &shl,const YsShellVertexHandle edVtHd[2],YSSIZE_T nVtInsert,const YsShellVertexHandle vtInsert[])
{
	return MakeInfo(shl,edVtHd[0],edVtHd[1],nVtInsert,vtInsert);
}

YSRESULT YsShellExt_EdgeVertexInsertionInfo::MakeInfo(const YsShellExt &shl,YsShellVertexHandle edVtHd0,YsShellVertexHandle edVtHd1,YSSIZE_T nVtInsert,const YsShellVertexHandle vtInsert[])
{
	plChange.CleanUp();

	auto edPlHd=shl.FindPolygonFromEdgePiece(edVtHd0,edVtHd1);
	for(auto plHd : edPlHd)
	{
		auto plVtHd=shl.GetPolygonVertex(plHd);

		for(auto idx : plVtHd.AllIndex())
		{
			if(plVtHd[idx]==edVtHd0 && plVtHd.GetCyclic(idx+1)==edVtHd1)
			{
				for(YSSIZE_T j=0; j<nVtInsert; ++j)
				{
					plVtHd.Insert(idx+1+j,vtInsert[j]);
				}
				break;
			}
			else if(plVtHd[idx]==edVtHd1 && plVtHd.GetCyclic(idx+1)==edVtHd0)
			{
				for(YSSIZE_T j=0; j<nVtInsert; ++j)
				{
					plVtHd.Insert(idx+1+j,vtInsert[nVtInsert-1-j]);
				}
				break;
			}
		}

		plChange.Increment();
		plChange.Last().plHd=plHd;
		plChange.Last().plVtHd.MoveFrom(plVtHd);
	}

	auto edCeHd=shl.FindConstEdgeFromEdgePiece(edVtHd0,edVtHd1);
	for(auto ceHd : edCeHd)
	{
		YSBOOL isLoop;
		YsArray <YsShellVertexHandle,16> ceVtHd;
		shl.GetConstEdge(ceVtHd,isLoop,ceHd);

		const YSSIZE_T idx1=(YSTRUE!=isLoop ? ceVtHd.GetN()-1 : ceVtHd.GetN());
		for(YSSIZE_T idx=0; idx<idx1; ++idx)
		{
			if(ceVtHd[idx]==edVtHd0 && ceVtHd.GetCyclic(idx+1)==edVtHd1)
			{
				for(YSSIZE_T j=0; j<nVtInsert; ++j)
				{
					ceVtHd.Insert(idx+1+j,vtInsert[j]);
				}
				break;
			}
			else if(ceVtHd[idx]==edVtHd1 && ceVtHd.GetCyclic(idx+1)==edVtHd0)
			{
				for(YSSIZE_T j=0; j<nVtInsert; ++j)
				{
					ceVtHd.Insert(idx+1+j,vtInsert[nVtInsert-1-j]);
				}
				break;
			}
		}

		ceChange.Increment();
		ceChange.Last().ceHd=ceHd;
		ceChange.Last().ceVtHd.MoveFrom(ceVtHd);
		ceChange.Last().isLoop=isLoop;
	}

	return YSOK;
}

const YsArray <YsShell_LocalOperation::PolygonChange> &YsShellExt_EdgeVertexInsertionInfo::PolygonChange(void) const
{
	return plChange;
}

const YsArray <YsShell_LocalOperation::ConstEdgeChange> &YsShellExt_EdgeVertexInsertionInfo::ConstEdgeChange(void) const
{
	return ceChange;
}

YSRESULT YsShell_SwapInfo::GeometryCheck(const YsShell &shl) const
{
	const YsVec3 edVtPos[2][2]=
	{
		{shl.GetVertexPosition(orgDiagonal[0]),shl.GetVertexPosition(orgDiagonal[1])},
		{shl.GetVertexPosition(newDiagonal[0]),shl.GetVertexPosition(newDiagonal[1])},
	};

	YsVec3 n[2];
	if(YSOK==YsGetNearestPointOfTwoLine(n[0],n[1],edVtPos[0][0],edVtPos[0][1],edVtPos[1][0],edVtPos[1][1]))
	{
		if(YSTRUE==YsCheckInBetween3(n[0],edVtPos[0]) &&
		   YSTRUE==YsCheckInBetween3(n[1],edVtPos[1]) &&
		   n[0]!=edVtPos[0][0] &&
		   n[0]!=edVtPos[0][1])
		{
			return YSOK;
		}
	}
	else
	{
		// Observed failure case:
		// (-4.307513  -2.496909   1.030980)-(-4.309643  -2.492926   0.009232) : Distances from newDiagonal point to the original diagonal 0.000699 0.001598
		// Distances are much greater than the tolerance.
		// Bad news: Merging bubbles to edge before this point won't help.

		const YsVec3 dgnVec[2]=
		{
			YsUnitVector(edVtPos[0][1]-edVtPos[0][0]),
			YsUnitVector(edVtPos[1][1]-edVtPos[1][0])
		};
		YsVec3 nom=dgnVec[0]^dgnVec[1];
		if(YSOK==nom.Normalize())
		{
			YsMatrix3x3 toXY(YSFALSE);
			toXY.MakeToXY(nom);

			const YsVec2 prjQuad[4]=
			{
				(toXY*shl.GetVertexPosition(quadVtHd[0])).xy(),
				(toXY*shl.GetVertexPosition(quadVtHd[1])).xy(),
				(toXY*shl.GetVertexPosition(quadVtHd[2])).xy(),
				(toXY*shl.GetVertexPosition(quadVtHd[3])).xy()
			};
			if(YSTRUE==YsCheckConvex2(4,prjQuad))
			{
				return YSOK;
			}
		}
	}
	return YSERR;
}
