/* ////////////////////////////////////////////////////////////

Polygon Crest
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

File Name: ysshellext_localop.h
//////////////////////////////////////////////////////////// */

#ifndef YSSHELLEXT_LOCALOP_IS_INCLUDED
#define YSSHELLEXT_LOCALOP_IS_INCLUDED
/* { */

#include <ysshellext.h>


class YsShell_LocalOperation
{
public:
	class PolygonChange
	{
	public:
		YsShellPolygonHandle plHd;
		YsArray <YsShellVertexHandle,4> plVtHd;
	};
	class ConstEdgeChange
	{
	public:
		YsShellExt::ConstEdgeHandle ceHd;
		YsArray <YsShellVertexHandle> ceVtHd;
		YSBOOL isLoop;
	};

	class EdgeSplit
	{
	public:
		YsShellVertexHandle edVtHd[2];
		YsVec3 pos;
		YsShellVertexHandle createdVtHd;

		inline void CleanUp(void)
		{
			edVtHd[0]=NULL;
			edVtHd[1]=NULL;
			createdVtHd=NULL;
		}
	};

	class PolygonSplit
	{
	public:
		YsShellPolygonHandle plHd;
		YsSegmentedArray <YsArray <YsShellVertexHandle,4>,4> plVtHdArray;
		YsSegmentedArray <YsArray <YsVec2,4>,4> prjVtPosArray;
	};
	class ConstEdgeSplit
	{
	public:
		YsShellExt::ConstEdgeHandle ceHd;
		YsSegmentedArray <YsArray <YsShellVertexHandle>,4> ceVtHdArray;
	};
	class FaceGroupSplit
	{
	public:
		YsShellExt::FaceGroupHandle fgHd;
		YsSegmentedArray <YsArray <YsShellPolygonHandle>,4> fgPlHdArray;

		void CleanUp(void);
	};
};


class YsShell_SwapInfo : public YsShell_LocalOperation
{
public:
	YsShellVertexHandle orgDiagonal[2],newDiagonal[2],quadVtHd[4];
	YsShellVertexHandle newTriVtHd[2][3];
	YsShellPolygonHandle triPlHd[2];

	YsShell_SwapInfo();
	void CleanUp(void);
	YSRESULT MakeInfo(const YsShell &shl,const YsShell::Edge edge);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellVertexHandle edVtHd[2]);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellVertexHandle edVtHd0,const YsShellVertexHandle edVtHd1);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellPolygonHandle edPlHd[2]);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellPolygonHandle edPlHd0,const YsShellPolygonHandle edPlHd1);

	/* Call this function after MakeInfo.  It returns YSERR if the nearest point of the extensions of the two edges, orgDiagonal and newDiagonal, is outside of either or both of the two edges. 
	   If the nearest point on the new diagonal is close to one of the points of the new diagonal, it means that a very flat triangle is being swapped away and is acceptable.
	   However, if the nearest point on the old diagonal is close to one of the points of the old diagonal, it means that a very flat triangle is being created.  
	   Therefore, in this case, the function returns YSERR. 
	   
       Acceptable.                       Unacceptable
	       |                               |
	       |                               |
	       |                               |
	      -+----New---------              -+----Old---------
	       |                               |
	       |                               |
	       |                               |
	   
	    */
	YSRESULT GeometryCheck(const YsShell &shl) const;
};


class YsShell_MergeInfo : public YsShell_LocalOperation
{
public:
	YsShellPolygonHandle plHd[2];
	YsArray <YsShellVertexHandle> newPlVtHd;
	YSBOOL duplicateWarning; // YSTRUE -> New polygon may be using a same vertex more than once.

	YsShell_MergeInfo();
	void CleanUp(void);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellPolygonHandle plHd[2]);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellPolygonHandle plHd0,const YsShellPolygonHandle plHd1);
};

class YsShell_CollapseInfo : public YsShell_LocalOperation
{
public:
	YsSegmentedArray <PolygonChange,4> newPlg;
	YsSegmentedArray <ConstEdgeChange,4> newConstEdge;
	YsArray <YsShellPolygonHandle> plHdToDel;
	YsArray <YsShellExt::ConstEdgeHandle> ceHdToDel;
	YsShellVertexHandle vtHdToDel,vtHdToSurvive;

	YsShell_CollapseInfo();
	void CleanUp(void);
	YSRESULT MakeInfo(const YsShellExt &shl,const YsShellVertexHandle vtHd[2]);
	YSRESULT MakeInfo(const YsShellExt &shl,const YsShellVertexHandle vtHdToDel,const YsShellVertexHandle vtHdToSurvive);
};

/*! This class makes a list of polygon modifications needed for inserting multiple vertices on an edge. */
class YsShellExt_EdgeVertexInsertionInfo : public YsShell_LocalOperation
{
private:
	YsArray <PolygonChange> plChange;
	YsArray <ConstEdgeChange> ceChange;

public:

	/*! Makes the insertion info.  The vertices to be inserted (vtInset[0]...vtInsert[nVtInsert-1]) must be sorted in the ascending order of the distance from edVtHd[0].
	    Or, some polygon edges may go backward. */
	YSRESULT MakeInfo(const YsShellExt &shl,const YsShellVertexHandle edVtHd[2],YSSIZE_T nVtInsert,const YsShellVertexHandle vtInsert[]);

	/*! Makes the insertion info.  The vertices to be inserted (vtInset[0]...vtInsert[nVtInsert-1]) must be sorted in the ascending order of the distance from edVtHd0.
	    Or, some polygon edges may go backward. */
	YSRESULT MakeInfo(const YsShellExt &shl,YsShellVertexHandle edVtHd0,YsShellVertexHandle edVtHd1,YSSIZE_T nVtInsert,const YsShellVertexHandle vtInsert[]);

	/*! Makes the insertion info.  The vertices to be inserted (vtInset[0]...vtInsert[nVtInsert-1]) must be sorted in the ascending order of the distance from edVtHd[0].
	    Or, some polygon edges may go backward. */
	template <const int N>
	YSRESULT MakeInfo(const YsShellExt &shl,const YsShellVertexHandle edVtHd[2],const YsArray <YsShellVertexHandle,N> &vtInsert)
	{
		return MakeInfo(shl,edVtHd,vtInsert.GetN(),vtInsert);
	}

	/*! Makes the insertion info.  The vertices to be inserted (vtInset[0]...vtInsert[nVtInsert-1]) must be sorted in the ascending order of the distance from edVtHd0.
	    Or, some polygon edges may go backward. */
	template <const int N>
	YSRESULT MakeInfo(const YsShellExt &shl,YsShellVertexHandle edVtHd0,YsShellVertexHandle edVtHd1,const YsArray <YsShellVertexHandle,N> &vtInsert)
	{
		return MakeInfo(shl,edVtHd0,edVtHd1,vtInsert.GetN(),vtInsert);
	}

	/*! Returns polygon changes. */
	const YsArray <PolygonChange> &PolygonChange(void) const;

	/*! Returns const-edge changes. */
	const YsArray <ConstEdgeChange> &ConstEdgeChange(void) const;
};

/*! This function inserts a vertex on an edge of the vertex array. */
template <class VertexType,const int N>
YSRESULT YsShellExt_InsertVertexOnEdgeOfVertexArray(
    YsArray <VertexType,N> &vtHdArray,YSBOOL isLoop,const VertexType edVtHd[2],VertexType vtHd)
{
	if(edVtHd[0]==vtHd || edVtHd[1]==vtHd || 2>vtHdArray.GetN())
	{
		return YSERR;
	}

	YSRESULT found=YSERR;
	for(YSSIZE_T idx=0; idx<vtHdArray.GetN()-1; ++idx)
	{
		if(YSTRUE==YsSameEdge(vtHdArray[idx],vtHdArray[idx+1],edVtHd[0],edVtHd[1]))
		{
			vtHdArray.Insert(idx+1,vtHd);
			found=YSOK;
		}
	}

	if(YSTRUE==isLoop && YSTRUE==YsSameEdge(vtHdArray.GetTop(),vtHdArray.GetEnd(),edVtHd[0],edVtHd[1]))
	{
		vtHdArray.Append(vtHd);
		found=YSOK;
	}

	return found;
}


/*! This function splits a loop into two loops at index idx1 and idx2.  
    The original (input) loop is loopVtHd0.  
    After split, the loop becomes two loops and will be stored in loopVtHd0 and loopVtHd1. 

	If no error, this function returns YSOK.

    Absolute value of idx0-idx1 must be at least 2, or this function returns YSERR.

    ****!  I'm sure I have written an exactly same function before.  I cannot just find it.
    */
template <class VertexType,const int N>
YSRESULT YsShellExt_SplitLoop(YsArray <VertexType,N> &loopVtHd0,YsArray <VertexType,N> &loopVtHd1,YSSIZE_T idx0,YSSIZE_T idx1)
{
	if(YSTRUE==loopVtHd0.IsInRange(idx0) && YSTRUE==loopVtHd0.IsInRange(idx1))
	{
		if(idx1<idx0)
		{
			YsSwapSomething (idx1,idx0);
		}

		if(2>idx1-idx0 || 2>loopVtHd0.GetN()-(idx1-idx0))
		{
			return YSERR;
		}

		loopVtHd1.Set(idx1-idx0+1,loopVtHd0.GetArray()+idx0);

		const YSSIZE_T toDel=idx1-idx0-1;
		loopVtHd0.Delete(idx0+1,toDel);

		return YSOK;
	}
	return YSERR;
}


/*! This function splits an open chain, chain0, at index of idx0 into two separate chains. 

	The first of the two will be returned in chain0, and the rest in chain1.
    */
template <class VertexType,const int N>
YSRESULT YsShellExt_SplitOpenChain(YsArray <VertexType,N> &chain0,YsArray <VertexType,N> &chain1,YSSIZE_T idx0)
{
	if(1<=idx0 && idx0<chain0.GetN()-1)
	{
		chain1.Set(chain0.GetN()-idx0,chain0.GetArray()+idx0);
		chain0.Resize(idx0+1);
		return YSOK;
	}
	return YSERR;
}

template <class VertexType,const int N>
YSRESULT YsShellExt_SplitLoopByHandle(YsArray <VertexType,N> &loopVtHd0,YsArray <VertexType,N> &loopVtHd1,VertexType vtx0,VertexType vtx1)
{
	YSSIZE_T idx0=-1,idx1=-1;
	for(auto idx : loopVtHd0.AllIndex())
	{
		if(loopVtHd0[idx]==vtx0)
		{
			idx0=idx;
		}
		if(loopVtHd0[idx]==vtx1)
		{
			idx1=idx;
		}
	}

	if(0<=idx0 && 0<=idx1)
	{
		return YsShellExt_SplitLoop(loopVtHd0,loopVtHd1,idx0,idx1);
	}
	return YSERR;
}

/* } */
#endif
