/* ////////////////////////////////////////////////////////////

Polygon Crest
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

File Name: ysshellext_sewingutil.cpp
//////////////////////////////////////////////////////////// */

#include "ysshellext_sewingutil.h"



void YsShellExt_SewingInfo::CleanUp(void)
{
	vtxOnEdge.CleanUp();
	vtxSequence.CleanUp();
	plgSplit.CleanUp();
}

YSRESULT YsShellExt_SewingInfo::MakeInfo(const YsShellExt &shl,YsShellVertexHandle vtHdFrom,YsShellVertexHandle vtHdTo)
{
	CleanUp();


	auto startDir=YsUnitVector(shl.GetVertexPosition(vtHdTo)-shl.GetVertexPosition(vtHdFrom));
	auto goalPos=shl.GetVertexPosition(vtHdTo);


	YsShellCrawler crawler;
	crawler.crawlingMode=YsShellCrawler::CRAWL_TO_GOAL;
	crawler.goal=shl.GetVertexPosition(vtHdTo);
	if(YSOK==crawler.Start(shl.Conv(),shl.GetVertexPosition(vtHdFrom),startDir,vtHdFrom))
	{
		vtxSequence.Increment();
		vtxSequence.Last().vtHd=vtHdFrom;
		vtxSequence.Last().plHd=crawler.currentPlHd;

		while(YSOK==crawler.Crawl(shl.Conv(),0.0))
		{
			if(crawler.currentState==YsShellCrawler::STATE_ON_VERTEX && crawler.currentEdVtHd[0]==vtxSequence.Last().vtHd)
			{
printf("%s %d\n",__FUNCTION__,__LINE__);
				vtxSequence.Last().plHd=crawler.currentPlHd;
			}
			else
			{
				switch(crawler.currentState)
				{
				case YsShellCrawler::STATE_IN_POLYGON:
					if(crawler.currentPos==goalPos) // Can happen in the end.
					{
						vtxSequence.Increment();
						vtxSequence.Last().plHd=crawler.currentPlHd;
						vtxSequence.Last().vtHd=vtHdTo;
					}
					break;
				case YsShellCrawler::STATE_ON_VERTEX:
					vtxSequence.Increment();
					vtxSequence.Last().plHd=crawler.currentPlHd;
					vtxSequence.Last().vtHd=crawler.currentEdVtHd[0];
					break;
				case YsShellCrawler::STATE_ON_EDGE:
					if(crawler.currentEdVtHd[0]==crawler.currentEdVtHd[1])
					{
						vtxSequence.Increment();
						vtxSequence.Last().plHd=crawler.currentPlHd;
						vtxSequence.Last().vtHd=crawler.currentEdVtHd[0];
					}
					else
					{
						vtxSequence.Increment();
						vtxSequence.Last().plHd=crawler.currentPlHd;
						vtxSequence.Last().vtHd=NULL;
						vtxSequence.Last().idx=vtxOnEdge.GetN();
						vtxOnEdge.Increment();
						vtxOnEdge.Last().edVtHd[0]=crawler.currentEdVtHd[0];
						vtxOnEdge.Last().edVtHd[1]=crawler.currentEdVtHd[1];
						vtxOnEdge.Last().createdVtHd=NULL;
						vtxOnEdge.Last().pos=crawler.currentPos;
					}
					break;
				}
			}

			if(YSTRUE==crawler.reachedDeadEnd || goalPos==crawler.currentPos)
			{
				break;
			}
		}

		if(crawler.currentPos==goalPos)
		{
			if(NULL==vtxSequence.Last().vtHd)
			{
				if(0<=vtxSequence.Last().idx)
				{
					printf("%d %d\n",shl.GetSearchKey(vtxOnEdge[vtxSequence.Last().idx].edVtHd[0]),shl.GetSearchKey(vtxOnEdge[vtxSequence.Last().idx].edVtHd[1]));
					printf("Landed between %s %s\n",
					   shl.GetVertexPosition(vtxOnEdge[vtxSequence.Last().idx].edVtHd[0]).Txt(),
					   shl.GetVertexPosition(vtxOnEdge[vtxSequence.Last().idx].edVtHd[1]).Txt());
				}
				else
				{
					printf("Nowhere to go.\n");
				}
			}
printf("%s %d\n",__FUNCTION__,__LINE__);
			return YSOK;
		}
	}
printf("%s %d\n",__FUNCTION__,__LINE__);
	return YSERR;
}

YsShellExt_SewingInfo::AllVtxOnEdgeMask YsShellExt_SewingInfo::AllVtxOnEdge(void)
{
	AllVtxOnEdgeMask mask;
	mask.vtxOnEdge=&vtxOnEdge;
	return mask;
}

void YsShellExt_SewingInfo::MakePolygonSplitInfo(const YsShellExt &shl)
{
	// A concave polygon may be cut twice by one crawling.
	YsShellPolygonAttribTable <YsArray <YsShell::Edge> > plHdToDivider(shl.Conv());

	for(auto &v : vtxSequence)
	{
		if(NULL==v.vtHd)
		{
			v.vtHd=vtxOnEdge[v.idx].createdVtHd;
		}
	}

	for(YSSIZE_T idx=0; idx<vtxSequence.GetN()-1; ++idx)
	{
		YsShell::Edge divider;
		divider.edVtHd[0]=vtxSequence[idx  ].vtHd;
		divider.edVtHd[1]=vtxSequence[idx+1].vtHd;

		auto plHd=vtxSequence[idx].plHd;

		auto *pldv=plHdToDivider[plHd];
		if(NULL==pldv)
		{
			YsArray <YsShell::Edge> newPlDv;
			newPlDv.Append(divider);
			plHdToDivider.SetAttrib(plHd,newPlDv);
		}
		else
		{
			pldv->Append(divider);
		}
	}

	for(auto hashHd : plHdToDivider.AllHandle())
	{
		auto plKey=plHdToDivider.GetKey(hashHd);
		auto plHd=shl.FindPolygon(plKey);
		auto &divider=*plHdToDivider[plHd];

		YsSegmentedArray <YsArray <YsShellVertexHandle,4>,4> plVtHd(1,NULL);
		plVtHd[0]=shl.GetPolygonVertex(plHd);

for(auto vtHd : plVtHd[0])
{
	printf("(%d)",shl.GetSearchKey(vtHd));
}
printf("\n");

		for(auto d : divider)
		{
printf("[%d %d]\n",shl.GetSearchKey(d.edVtHd[0]),shl.GetSearchKey(d.edVtHd[1]));
			YsArray <YsShellVertexHandle,4> fragment;
			for(auto &pl : plVtHd)
			{
				if(YSOK==YsShellExt_SplitLoopByHandle(pl,fragment,d.edVtHd[0],d.edVtHd[1]))
				{
printf("%s %d\n",__FUNCTION__,__LINE__);
					plVtHd.Append(fragment);
					break;
				}
			}
		}

		plgSplit.Increment();
		plgSplit.Last().plHd=plHd;
		plgSplit.Last().plVtHdArray=plVtHd;
	}
}

YsShellExt_SewingInfo::AllPolygonSplitMask YsShellExt_SewingInfo::AllPolygonSplit(void) const
{
	AllPolygonSplitMask mask;
	mask.plgSplitPtr=&plgSplit;
	return mask;
}
